# -*- coding: utf-8 -*-
import datetime
import bisect
import re
import warnings
from copy import deepcopy

import six
import pandas as pd

from rqdatac.client import get_client
from rqdatac.utils import to_date, datetime_to_int14, to_date_str, str_to_dt_time, int8_to_date
from rqdatac.validators import (
    ensure_list_of_string,
    ensure_date_int,
    check_type,
    ensure_date_str,
    ensure_order_book_id,
    ensure_order_book_ids,
    check_items_in_container,
    ensure_date_range,
    ensure_int,
    ensure_string,
    ensure_date_or_today_int,
    ensure_string_in,
)
from rqdatac.services.concept import concept_names as get_concept_names
from rqdatac.services.shenwan import get_instrument_industry
from rqdatac.services.constant import SectorCode, SectorCodeItem, IndustryCode, IndustryCodeItem
from rqdatac.services.calendar import get_previous_trading_date, is_trading_date, has_night_trading
from rqdatac.decorators import export_as_api, ttl_cache, compatible_with_parm
from dateutil.relativedelta import relativedelta


def _id_convert_one(order_book_id):  # noqa: C901
    # hard code
    if order_book_id in {"T00018", "T00018.SH", "T00018.XSHG", "SH.T00018"}:
        return "990018.XSHG"

    if order_book_id.isdigit():
        if order_book_id.startswith("0") or order_book_id.startswith("3"):
            return order_book_id + ".XSHE"
        elif (
                order_book_id.startswith("5")
                or order_book_id.startswith("6")
                or order_book_id.startswith("9")
                or order_book_id.startswith("15")
        ):
            return order_book_id + ".XSHG"
        else:
            raise ValueError("order_book_ids should be str like 000001, 600000")
    order_book_id = order_book_id.upper()
    if order_book_id.endswith(".XSHG") or order_book_id.endswith(".XSHE"):
        return order_book_id

    if order_book_id.startswith("SZ"):
        return order_book_id.replace(".", "")[2:] + ".XSHE"
    elif order_book_id.startswith("SH"):
        return order_book_id.replace(".", "")[2:] + ".XSHG"
    elif order_book_id.endswith("SZ"):
        return order_book_id.replace(".", "")[:-2] + ".XSHE"
    elif order_book_id.endswith("SH"):
        return order_book_id.replace(".", "")[:-2] + ".XSHG"

    # 期货
    order_book_id = order_book_id.replace("-", "").split(".")[0]
    try:
        res = re.findall(r"^([A-Z]+)(\d+)([PC]\w+)?", order_book_id)[0]
    except IndexError:
        raise ValueError("unknown order_book_id: {}".format(order_book_id))
    if len(res[1]) == 3 and res[1] != '888':
        year = str(datetime.datetime.now().year + 3)
        # 按照当前年份+3之后的年份取十位位置上数字递减重新组合为 order_book_id列表 去查询 trading_code 相等的最大合约(即返回结果 trading_code 相等的首个合约)
        # bug：当 year[-2] = 0(2097年)，若(设输入为'TA312')没有查到此 order_book_id('TA0312')， 则不会对三位代码补位而直接返回原 trading_code('TA312')
        ins_infos = instruments([res[0] + str(n) + res[1] + res[2] for n in range(int(year[-2]), -1, -1)])
        for ins_info in ins_infos:
            if ins_info is None:
                continue
            trading_code = getattr(ins_info, "trading_code", "")
            if trading_code.upper() == order_book_id.upper():
                return ins_info.order_book_id
    return order_book_id


@export_as_api
def id_convert(order_book_ids):
    """合约格式转换

    :param order_book_ids: str 或 str list, 如'000001', 'SZ000001', '000001SZ',
        '000001.SZ', 纯数字str默认为股票类型
    :returns: str 或 str list, 米筐格式的合约

    """
    if isinstance(order_book_ids, six.string_types):
        return _id_convert_one(order_book_ids)
    elif isinstance(order_book_ids, list):
        return [_id_convert_one(o) for o in order_book_ids]
    else:
        raise ValueError("order_book_ids should be str or list")


def _id_compatible(order_book_id):
    if order_book_id.endswith("XSHE"):
        return order_book_id[:-4] + "SZ"
    elif order_book_id.endswith("XSHG"):
        return order_book_id[:-4] + "SH"
    else:
        return order_book_id


def _all_instruments_list(market):
    ins = [Instrument(i) for i in get_client().execute("all_instruments", market=market)]

    extra_hk_instruments = []
    if market == 'hk':  # 对港股需要根据 stock_connect 字段做拆分并替换该字段
        suffix_map = {'sz': 'XSEC', 'sh': 'XSSC'}
        for i in ins:
            if not i.stock_connect:  # 港股中 stcok_connect 字段为空则设置该字段为 ''
                setattr(i, 'stock_connect', '')
                continue
            for stock_connect in i.stock_connect:
                _temp_instruments = deepcopy(i)
                _temp_instruments.unique_id = _temp_instruments.unique_id[:-4] + suffix_map.get(stock_connect)
                _temp_instruments.stock_connect = stock_connect + '_connect'
                extra_hk_instruments.append(_temp_instruments)
            i.stock_connect = i.stock_connect[0] if len(i.stock_connect) == 1 else '_and_'.join(i.stock_connect)

    ins += extra_hk_instruments
    return ins


@ttl_cache(3 * 3600)
def _all_cached_instruments_list(market):
    return _all_instruments_list(market)


@ttl_cache(3 * 3600)
def _all_instruments_dict(market):
    ins = _all_cached_instruments_list(market)
    result = dict()
    for i in ins:
        result[i.symbol] = i
        if i.type == "Convertible":
            result[_id_compatible(i.order_book_id)] = i

        if getattr(i, "unique_id", None):  # 对港股 unique_id 作为 key 添加到 result dict
            result[i.unique_id] = i

        if i.order_book_id in result:  # 对港股存在退市后 order_book_id 复用的情况只存上市日期最晚的信息
            if i.listed_date > result[i.order_book_id].listed_date:
                result[i.order_book_id] = i
        else:
            result[i.order_book_id] = i

    try:
        result["沪深300"] = result["000300.XSHG"]
        result["中证500"] = result["000905.XSHG"]
        result[result["SSE180.INDX"].symbol] = result["000010.XSHG"]
    except KeyError:
        pass

    return result


def get_underlying_listed_date(underlying_symbol, ins_type, market="cn"):
    """ 获取期货或者期权的某个品种的上市日期"""
    ins_list = _all_cached_instruments_list(market)
    listed_dates = [i.listed_date for i in ins_list
                    if (getattr(i, "underlying_symbol", "") == underlying_symbol
                        and i.type == ins_type and i.listed_date != "0000-00-00")]

    return min(listed_dates)


def get_tick_size(order_book_id, market="cn"):
    """获取合约价格最小变动单位

    :param order_book_id: 如: FU1703
    :param market: 如：'cn' (Default value = "cn")
    :returns: float

    """
    return get_client().execute("get_tick_size", order_book_id, market=market)


HK_STOCK_PRICE_SECTIONS = [0.25, 0.5, 10, 20, 100, 200, 500, 1000, 2000, 5000]
HK_STOCK_TICK_SIZES = [0.001, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5]


# noinspection All
class Instrument(object):
    def __init__(self, d):
        self.__dict__ = d

    def __repr__(self):
        if self.has_citics_info() and not hasattr(self, "_citics_industry_code"):
            self.citics_industry()

        return "{}({})".format(
            type(self).__name__,
            ", ".join(
                [
                    "{}={!r}".format(k.lstrip("_"), v)
                    for k, v in self.__dict__.items()
                    if v is not None
                ]
            ),
        )

    @property
    def concept_names(self):
        return get_concept_names(self.order_book_id)

    def days_from_listed(self, date=None):
        if self.listed_date == "0000-00-00":
            return -1

        date = to_date(date) if date else datetime.date.today()
        if self.de_listed_date != "0000-00-00" and date > to_date(self.de_listed_date):
            # 晚于退市日期
            return -1

        listed_date = to_date(self.listed_date)
        ipo_days = (date - listed_date).days
        return ipo_days if ipo_days >= 0 else -1

    def days_to_expire(self, date=None):
        if getattr(self, 'maturity_date', '0000-00-00') == '0000-00-00':
            return -1

        date = to_date(date) if date else datetime.date.today()

        maturity_date = to_date(self.maturity_date)
        days = (maturity_date - date).days
        return days if days >= 0 else -1

    def tick_size(self, price=None):
        if self.exchange == "XHKG":
            check_type(price, (int, float), "price")
            index = bisect.bisect_left(HK_STOCK_PRICE_SECTIONS, price)
            return HK_STOCK_TICK_SIZES[index]
        elif self.type in ["CS", "INDX"]:
            return 0.01
        elif self.type in ["ETF", "LOF", "FenjiB", "FenjiA", "FenjiMu"]:
            return 0.001
        elif self.type == "Convertible":
            if self.exchange == "XSHG":
                return 0.01
            else:
                return 0.001
        elif self.type not in ["Future", "Option", "Spot"]:
            return -1
        return get_tick_size(self.order_book_id)

    def has_citics_info(self):
        return self.type == "CS" and self.exchange in {"XSHE", "XSHG"}

    def citics_industry(self, date=None):
        if self.has_citics_info():
            if date is None:
                if hasattr(self, "_citics_industry_code"):
                    return (self._citics_industry_code, self._citics_industry_name)

            if self.de_listed_date != '0000-00-00':
                date = get_previous_trading_date(self.de_listed_date)

            result = get_instrument_industry(self.order_book_id, date=date, level=1, source='citics_2019')
            if result is None:
                self._citics_industry_code, self._citics_industry_name = (None, None)
                return None

            self._citics_industry_code = result['first_industry_code'][0]
            self._citics_industry_name = result['first_industry_name'][0]

            return {"code": result.iloc[0, 0], "name": result.iloc[0, 1]}

    @property
    def citics_industry_code(self):
        if not self.has_citics_info():
            return None

        if not hasattr(self, "_citics_industry_code"):
            self.citics_industry()
        return self._citics_industry_code

    @property
    def citics_industry_name(self):
        if not self.has_citics_info():
            return None

        if not hasattr(self, "_citics_industry_name"):
            self.citics_industry()
        return self._citics_industry_name


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def instruments(order_book_ids, market="cn"):
    """获取证券详细信息

    :param order_book_ids: 证券ID列表, 如'000001.XSHE', 'AAPL.US'. 注意, 所有列表中的证券需要属于同一个国家。
    :param market: 证券所属国家, 如 cn, us, hk (Default value = "cn")
    :returns: 对应证券的列表

    """

    all_dict = _all_instruments_dict(market)
    if isinstance(order_book_ids, six.string_types):
        try:
            return all_dict[order_book_ids]
        except KeyError:
            warnings.warn('unknown order_book_id: {}'.format(order_book_ids))
            return
    all_list = (all_dict.get(i) for i in order_book_ids)
    return [i for i in all_list if i]


VALID_TYPES = {"CS", "ETF", "LOF", "INDX", "Future", "Spot", "Option", "Convertible", "Repo"}


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def all_instruments(type=None, date=None, market="cn", **kwargs):
    """获得某个国家的全部证券信息

    :param type:  (Default value = None)
    :param date:  (Default value = None)
    :param market: cn, hk (Default value = "cn")
    :kwargs
        trading_market: [hk, all] (Default value = "hk")
            hk: 港交所可购买的股票。对应返回stock_connect = null、sh、sz 的记录
            all: 包括港交所、上交所、深交所可购买的港股。（对沪深港通支持股票均展示一条独立的unique_id捆绑的信息）,对应返回全部列表，即stock_connect = null、sz_and_sh、sh、sz、sz_connect、sh_connect
    """

    if type is None:
        itype = VALID_TYPES
    else:
        type = ensure_list_of_string(type)
        itype = set()
        for t in type:
            if t.upper() == "STOCK":
                itype.add("CS")
            elif t.upper() == "FUND":
                itype = itype.union({"ETF", "LOF"})
            elif t.upper() == "INDEX":
                itype.add("INDX")
            elif t not in VALID_TYPES:
                raise ValueError("invalid type: {}, chose any in {}".format(type, VALID_TYPES))
            else:
                itype.add(t)

    if date:
        date = ensure_date_str(date)
        cond = lambda x: (  # noqa: E731
                x.type in itype
                and (x.listed_date <= date or x.listed_date == "0000-00-00")
                and (
                        x.de_listed_date == "0000-00-00"
                        or (
                                x.de_listed_date >= date
                                and x.type in ("Future", "Option")
                                or (x.de_listed_date > date and x.type not in ("Future", "Option"))
                        )
                )
        )
    else:
        cond = lambda x: x.type in itype  # noqa: E731

    cached = kwargs.pop("cached", True)
    trading_market = kwargs.pop("trading_market", 'hk')
    if kwargs:
        raise ValueError("Unknown kwargs: {}".format(kwargs))

    if cached:
        get_instrument_list = _all_cached_instruments_list
    else:
        get_instrument_list = _all_instruments_list

    ins_ret = filter(cond, get_instrument_list(market))

    if market == 'hk' and trading_market == 'hk':
        ins_ret = filter(lambda x: x.unique_id.endswith('.XHKG'), ins_ret)

    if len(itype) == 1:
        df = pd.DataFrame([v.__dict__ for v in ins_ret])
        internal_fields = [f for f in df.columns if f.startswith('_')]
        for f in internal_fields:
            del df[f]
    else:
        df = pd.DataFrame(
            [
                (
                    v.order_book_id,
                    v.symbol,
                    getattr(v, "abbrev_symbol", None),
                    v.type,
                    v.listed_date,
                    v.de_listed_date,
                )
                for v in ins_ret
            ],
            columns=[
                "order_book_id",
                "symbol",
                "abbrev_symbol",
                "type",
                "listed_date",
                "de_listed_date",
            ],
        )
    return df


def to_sector_name(s):
    for _, v in SectorCode.__dict__.items():
        if isinstance(v, SectorCodeItem):
            if v.cn == s or v.en == s or v.name == s:
                return v.name
    return s


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def sector(code, market="cn"):
    """获取某个板块的股票列表。目前支持的板块分类具体可以查询以下网址:
    https://www.ricequant.com/api/research/chn#research-API-sector

    :param code: 可以使用板块英文名字如'Energy', 或者 sector_code.Energy
    :param market: 地区代码, 如'cn' (Default value = "cn")
    :returns: 板块全部股票列表

    """
    check_type(code, (SectorCodeItem, six.string_types), "code")
    if isinstance(code, SectorCodeItem):
        code = code.name
    else:
        code = to_sector_name(code)
    return [
        v.order_book_id
        for v in _all_cached_instruments_list(market)
        if v.type == "CS" and v.sector_code == code
    ]


def to_industry_code(s):
    for _, v in IndustryCode.__dict__.items():
        if isinstance(v, IndustryCodeItem):
            if v.name == s:
                return v.code
    return s


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def industry(code, market="cn"):
    """获取某个行业的股票列表。目前支持的行业列表具体可以查询以下网址:
    https://www.ricequant.com/api/research/chn#research-API-industry

    :param code: 行业代码,如 A01, 或者 industry_code.A01
    :param market: 地区代码, 如'cn' (Default value = "cn")
    :returns: 行业全部股票列表

    """
    if not isinstance(code, six.string_types):
        code = code.code
    else:
        code = to_industry_code(code)
    return [
        v.order_book_id
        for v in _all_cached_instruments_list(market)
        if v.type == "CS" and v.industry_code == code
    ]


@export_as_api
def get_future_contracts(underlying_symbol, date=None, market="cn"):
    import rqdatac
    import warnings

    msg = "'get_future_contracts' is deprecated, please use 'futures.get_contracts' instead"
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    return rqdatac.futures.get_contracts(underlying_symbol, date, market)


@export_as_api(namespace='futures')
def get_contracts(underlying_symbol, date=None, market="cn"):
    """获得中国市场某个时间某个期货品种正在交易的合约列表

    :param underlying_symbol: 期货品种
    :param date: 日期，可以为str，datetime，date，pd.Timestamp 等类型
    :param market:  (Default value = "cn")
    :returns: list of order book id

    """
    if date is None:
        date = datetime.date.today()
    date = ensure_date_str(date)

    return sorted(
        v.order_book_id
        for v in _all_cached_instruments_list(market)
        if v.type == "Future"
        and v.underlying_symbol == underlying_symbol
        and v.listed_date != "0000-00-00"
        and v.listed_date <= date <= v.de_listed_date
    )


@export_as_api
def jy_instrument_industry(order_book_ids, date=None, level=1, expect_df=True, market="cn"):
    """获取股票对应的聚源行业

    :param order_book_ids: 股票列表，如['000001.XSHE', '000002.XSHE']
    :param date: 如 '2015-01-07' (Default value = None)
    :param level: 聚源等级，0, 1, 2, 3, 'customized' (Default value = 1)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market:  (Default value = "cn")
    :returns: code, name
        返回输入日期最近交易日的股票对应聚源行业以及对应的聚源等级

    """
    if level not in (0, 1, 2, 3, "customized"):
        raise ValueError("level should in 0, 1, 2, 3, 'customized'")
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    if not date:
        date = ensure_date_int(get_previous_trading_date(datetime.date.today(), market=market))
    else:
        date = ensure_date_int(date)

    df = get_client().execute("jy_instrument_industry", order_book_ids, date, level, market=market)
    if not df:
        return
    if len(order_book_ids) == 1 and not expect_df:
        r = df[0]
        if level == 0:
            return r["first_industry_name"], r["second_industry_name"], r["third_industry_name"]
        return r["industry_name"]
    return pd.DataFrame(df).set_index("order_book_id")


@export_as_api(namespace="econ")
def get_reserve_ratio(reserve_type="all", start_date=None, end_date=None, market="cn"):
    """获取存款准备金率

    :param reserve_type: 存款准备金详细类别，默认为'all'，目前仅支持'all'、'major'、'other'类别的查询
    :param start_date: 开始查找时间，如'20180501'，默认为上一年的当天
    :param end_date: 截止查找时间，如'20180501'，默认为当天
    :param market:  (Default value = "cn")
    :return: pd.DataFrame

    """
    check_items_in_container(reserve_type, ["all", "major", "other"], "reserve_type")

    start_date, end_date = ensure_date_range(start_date, end_date, delta=relativedelta(years=1))

    ret = get_client().execute(
        "econ.get_reserve_ratio", reserve_type, start_date, end_date, market
    )
    if not ret:
        return
    columns = ["info_date", "effective_date", "reserve_type", "ratio_floor", "ratio_ceiling"]
    df = pd.DataFrame(ret, columns=columns).set_index("info_date").sort_index(ascending=True)
    return df


@export_as_api(namespace="econ")
def get_money_supply(start_date=None, end_date=None, market="cn"):
    """获取货币供应量信息

    :param start_date: 开始日期，默认为一年前
    :param end_date: 结束日期，默认为今天
    :param market:  (Default value = "cn")

    """
    check_items_in_container("info_date", ["info_date", "effective_date"], "date_type")
    start_date, end_date = ensure_date_range(start_date, end_date, delta=relativedelta(years=1))

    data = get_client().execute("econ.get_money_supply", start_date, end_date, market=market)
    if not data:
        return
    columns = [
        "info_date",
        "effective_date",
        "m2",
        "m1",
        "m0",
        "m2_growth_yoy",
        "m1_growth_yoy",
        "m0_growth_yoy",
    ]
    df = pd.DataFrame(data, columns=columns).set_index("info_date").sort_index(ascending=True)
    return df


@export_as_api
def get_main_shareholder(
        order_book_ids, start_date=None, end_date=None, is_total=False, start_rank=None, end_rank=None, market="cn"
):
    """获取十大股东信息

    :param order_book_ids: 股票代码
    :param start_date: 开始日期，默认为一年前
    :param end_date: 结束日期，默认为今天
    :param is_total: 是否十大股东, True 和 False，默认为False
    :param start_rank: 开始排名, int, 默认为最小排名
    :param end_rank: 结束排名, int, 默认为最大排名
    :param market:  (Default value = "cn")

    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    check_items_in_container(is_total, [True, False], "is_total")
    start_date, end_date = ensure_date_range(start_date, end_date, delta=relativedelta(years=1))

    ret = get_client().execute(
        "get_main_shareholder", order_book_ids, start_date, end_date, is_total,
        start_rank=start_rank, end_rank=end_rank, market=market
    )
    if not ret:
        return
    columns = ['info_date', 'end_date', 'rank', 'shareholder_name', 'shareholder_attr', 'shareholder_kind',
               'shareholder_type', 'hold_percent_total', 'hold_percent_float', 'share_pledge', 'share_freeze',
               'order_book_id']
    df = pd.DataFrame(ret, columns=columns).sort_values(by=['info_date', 'rank']).\
        set_index(['order_book_id', 'info_date'])
    return df


@export_as_api
def get_current_news(n=None, start_time=None, end_time=None, channels=None):
    """获取新闻
    :param n: 新闻条数, n 和 start_time/end_time 只能指定其一
    :param start_time: 开始日期，默认为None,格式为%Y-%m-%d %H:%M:%S，如"2018-10-20 09:10:20"
    :param end_time: 结束日期，默认为None,格式为%Y-%m-%d %H:%M:%S，如"2018-10-20 19:10:20"
    :param channels: 新闻大类, 默认为None,返回每个大类n条新闻, 如 "global"，"forex", "commodity", "a-stock"
    """

    if start_time is not None or end_time is not None:
        try:
            start_time = datetime.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
        except Exception:
            raise ValueError('start_time should be str format like "%Y-%m-%d %H:%M:%S"')
        try:
            end_time = datetime.datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S")
        except Exception:
            raise ValueError('end_time should be str format like "%Y-%m-%d %H:%M:%S"')
        start_time = datetime_to_int14(start_time)
        end_time = datetime_to_int14(end_time)
        if n is not None:
            raise ValueError(
                "please either specify parameter n, or specify both start_time and end_time"
            )
        n = 1200
    elif n is None:
        n = 1
    else:
        n = ensure_int(n, "n")
        if n < 1 or n > 1200:
            raise ValueError("n should be in [0, 1200]")

    if channels is not None:
        channels = ensure_list_of_string(channels)
        check_items_in_container(channels, ["global", "forex", "a-stock", "commodity"], "channels")
    else:
        channels = ["global", "forex", "a-stock", "commodity"]

    data = get_client().execute("get_current_news", n, start_time, end_time, channels)
    if not data:
        return
    df = pd.DataFrame(data, columns=["channel", "datetime", "content"])
    return df.set_index("channel")


@export_as_api(namespace="econ")
def get_factors(factors, start_date, end_date, market="cn"):
    start_date, end_date = ensure_date_range(start_date, end_date)
    factors = ensure_list_of_string(factors, "factors")
    data = get_client().execute("econ.get_factors", factors, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df = df.reindex(columns=["factor", "info_date", "start_date", "end_date", "value"])
    df.set_index(["factor", "info_date"], inplace=True)
    return df


@export_as_api
def get_trading_hours(order_book_id, date=None, expected_fmt="str", frequency="1m", market="cn"):
    """获取合约指定日期交易时间
      :param order_book_id: 合约代码
      :param date: 日期，默认为今天
      :param expected_fmt: 返回格式，默认为str, 也支持datetime.time和datetime.datetime格式
      :param frequency: 频率，默认为1m, 对应米筐分钟线时间段的起始, tick和1m相比区别在于每个交易时间段开盘往前移一分钟
      :param market:  (Default value = "cn")

      :return: trading_hours str or list of datetime.time/datetime.datetime list or None
      """
    date = ensure_date_or_today_int(date)
    if not is_trading_date(date, market):
        warnings.warn(" %d is not a trading date" % date)
        return

    ensure_string(order_book_id, "order_book_id")
    ins = instruments(order_book_id)
    if ins is None:
        return

    ensure_string_in(expected_fmt, ("str", "time", "datetime"), "expected_fmt")
    ensure_string_in(frequency, ("1m", "tick"), "frequency")
    date_str = to_date_str(date)

    if ins.listed_date > date_str:
        return

    if ins.type in ("Future", "Option") and ins.de_listed_date < date_str and ins.de_listed_date != "0000-00-00":
        return

    if ins.type not in ("Future", "Option") and ins.de_listed_date <= date_str and ins.de_listed_date != "0000-00-00":
        return
    if ins.type == "Repo":
        trading_hours = "09:31-11:30,13:01-15:30"
    elif ins.type == "Spot":
        if has_night_trading(date, market):
            trading_hours = "20:01-02:30,09:01-15:30"
        else:
            trading_hours = "09:01-15:30"
    elif ins.type not in ("Future", "Option") or (ins.type == "Option" and ins.exchange in ("XSHG", "XSHE")):
        trading_hours = "09:31-11:30,13:01-15:00"
    else:
        trading_hours = get_client().execute("get_trading_hours", ins.underlying_symbol, date, market=market)
        if trading_hours is None:
            return
        # 前一天放假或者该品种上市首日没有夜盘
        no_night_trading = (not has_night_trading(date, market) or
                            get_underlying_listed_date(ins.underlying_symbol, ins.type) == date_str)

        if no_night_trading and not trading_hours.startswith("09"):
            trading_hours = trading_hours.split(",", 1)[-1]

    if frequency == "tick":
        trading_hours = ",".join([s[:4] + str(int(s[4]) - 1) + s[5:] for s in trading_hours.split(",")])

    if expected_fmt != "str":
        trading_hours = [t.split("-", 1) for t in trading_hours.split(",")]
        for i, (start, end) in enumerate(trading_hours):
            trading_hours[i][0] = str_to_dt_time(start)
            trading_hours[i][1] = str_to_dt_time(end)

        if expected_fmt == "datetime":
            td = int8_to_date(date)
            prev_td = get_previous_trading_date(date)
            prev_td_next = prev_td + datetime.timedelta(days=1)

            for i, (start, end) in enumerate(trading_hours):
                if start.hour > 16:
                    start_dt = prev_td
                    end_dt = start_dt if end.hour > 16 else prev_td_next
                else:
                    start_dt = end_dt = td
                trading_hours[i][0] = datetime.datetime.combine(start_dt, start)
                trading_hours[i][1] = datetime.datetime.combine(end_dt, end)

    return trading_hours


@export_as_api
def get_private_placement(order_book_ids, start_date=None, end_date=None, progress="complete", issue_type="private", market="cn"):
    """获取定增数据
    :param order_book_ids: 合约代码
    :param start_date: 开始日期，默认为None
    :param end_date: 结束日期，默认为None
    :param progress: 是否已完成定增，默认为complete。可选参数["complete", "incomplete", "all"]
    :param issue_type: 默认为all。可选参数["private", "public", "all"]
    :param market: (Default value = "cn")
    :return:
    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    if start_date and end_date:
        start_date, end_date = ensure_date_range(start_date, end_date)
    elif start_date:
        start_date = ensure_date_int(start_date)
    elif end_date:
        end_date = ensure_date_int(end_date)
    ensure_string_in(progress, ["complete", "incomplete", "all"], "progress")
    ensure_string_in(issue_type, ["private", "public", "all"], "issue_type")
    issue_type_change = {"private": (21, 23), "public": (22,), "all": (21, 22, 23)}
    issue_type = issue_type_change[issue_type]
    data = get_client().execute(
        "get_private_placement", order_book_ids, start_date, end_date, progress, issue_type=issue_type, market=market
    )
    if not data:
        return
    progress_map = {
        10: "董事会预案", 20: "股东大会通过", 21: "国资委通过", 22: "发审委通过", 23: "证监会通过",
        29: "实施中", 30: "实施完成", 40: "国资委否决", 41: "股东大会否决", 42: "证监会否决",
        43: "发审委否决", 50: "延期实施", 60: "停止实施", 70: "暂缓发行"}
    issue_type_map = {21: "非公开发行", 22: "公开发行", 23: "非公开发行配套融资"}
    df = pd.DataFrame(data)
    df["progress"] = df["progress"].map(progress_map)
    df["issue_type"] = df["issue_type"].map(issue_type_map)
    df.set_index(["order_book_id", "initial_info_date"], inplace=True)
    return df


@export_as_api
def get_share_transformation(predecessor=None, market="cn"):
    """
    获取转股信息
    :param predecessor: 换股前的股票代码。默认为空，返回所有转股信息
    :param market:  (Default value = "cn")
    :return pd.DataFrame
    """
    if predecessor:
        predecessor = ensure_order_book_id(predecessor)
    data = get_client().execute("get_share_transformation", predecessor, market=market)
    if not data:
        return
    columns = [
        "predecessor", "successor", "effective_date", "share_conversion_ratio", "predecessor_delisted",
        "discretionary_execution", "predecessor_delisted_date", "event"
    ]
    df = pd.DataFrame(data, columns=columns)
    df = df.sort_values('predecessor').reset_index(drop=True)
    return df


@export_as_api(namespace="user")
def get_quota():
    """
    获取用户配额信息
    :return dict
        bytes_limit：每日流量使用上限（单位为字节），如为0则表示不受限
        bytes_used：当日已用流量（单位为字节）
        remaining_days：账号剩余有效天数, 如为0则表示不受限
        license_type：账户类型(FULL: 付费类型，TRIAL: 试用类型， EDU: 教育网类型, OTHER: 其他类型)
    """
    data = get_client().execute("user.get_quota")
    if data['bytes_limit'] > 0 and data["bytes_used"] >= data["bytes_limit"]:
        warnings.warn("Traffic usage has been exceeded quota,"
                      "Please call us at 0755-22676337 to upgrade"
                      "your contract.")
    return data


_CHECK_CATEGORIES = ("stock_1d", "stock_1m", "future_1d", "future_1m", "index_1d", "index_1m")


@export_as_api()
def get_update_status(categories):
    """
    获取数据最新日期
    :param categories: str or list or str, 数据类型，支持类型有:
        stock_1d: 股票日线
        stock_1m: 股票分钟线
        future_1d: 期货日线
        future_1m: 期货分钟线
        index_1d：指数日线
        index_1m：指数分钟线

    :return datetime.datetime or dict(category=datetime.datetime)
    """
    categories = ensure_list_of_string(categories, "categories")
    check_items_in_container(categories, _CHECK_CATEGORIES, "categories")
    ret = get_client().execute("get_update_status", categories)
    if len(categories) == 1:
        return ret[0]["date"]
    return {r["category"]: r["date"] for r in ret}


@export_as_api()
def info():
    """
    打印账户信息
    :return None
    """
    get_client().info()


@export_as_api()
def get_basic_info(order_book_ids=None, fields=("order_book_id", "symbol"), market='cn'):
    if order_book_ids is not None:
        order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    if fields is not None:
        fields = ensure_list_of_string(fields, "fields")

    ret = get_client().execute("get_basic_info", order_book_ids, fields, market=market)
    if not ret:
        return
    columns, data = ret
    return pd.DataFrame(data, columns=columns)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
import numpy as np
import pandas as pd
from rqdatac.services.calendar import (get_previous_trading_date,
                                       is_trading_date)

FIELD_METHOD_MAP = {
    "open": "first",
    "close": "last",
    "iopv": "last",
    "high": np.maximum,
    "low": np.minimum,
    "limit_up": np.maximum,
    "limit_down": np.minimum,
    "total_turnover": np.add,
    "volume": np.add,
    "num_trades": np.add,
    "acc_net_value": "last",
    "unit_net_value": "last",
    "discount_rate": "last",
    "settlement": "last",
    "prev_settlement": "last",
    "open_interest": "last",
    "basis_spread": "last",
    "date": "last",
    "trading_date": "last",
    "datetime": "last",
}
FIELD_METHOD_MAP2 = {
    "open": "first",
    "close": "last",
    "iopv": "last",
    "high": "max",
    "low": "min",
    "total_turnover": "sum",
    "volume": "sum",
    "num_trades": "sum",
    "acc_net_value": "last",
    "unit_net_value": "last",
    "discount_rate": "last",
    "settlement": "last",
    "prev_settlement": "last",
    "open_interest": "last",
    "basis_spread": "last",
    "contract_multiplier": "last",
    "strike_price": "last",
}


def _update_weekly_trading_date_index(idx):
    if is_trading_date(idx[1]):
        return idx
    return idx[0], get_previous_trading_date(idx[1])


def resample_week_df(df, fields):
    hows = {field: FIELD_METHOD_MAP2[field] for field in fields}
    res1 = df[df['volume'] > 0]
    if not res1.empty:
        res1 = res1.groupby(level='order_book_id').resample('W-Fri', level=1).agg(hows)
    res2 = df[df['volume'] == 0]
    if not res2.empty:
        res2 = res2.groupby(level='order_book_id').resample('W-Fri', level=1).agg(hows)
    res = pd.concat([res1, res2])
    res.index = res.index.map(_update_weekly_trading_date_index)
    res = res[~res.index.duplicated(keep='first')]
    res.sort_index(inplace=True)
    return res



# -*- coding: utf-8 -*-
import datetime

import pandas as pd

from rqdatac.validators import (
    check_items_in_container,
    ensure_date_int,
    ensure_order_book_ids,
    ensure_string,
    ensure_string_in,
    ensure_date_range,
)

from rqdatac import get_trading_dates
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api
from rqdatac.services.basic import all_instruments, instruments
from rqdatac.services.get_price import get_price
from rqdatac.utils import is_panel_removed

VALID_GREEKS_FIELDS = ['iv', 'delta', 'gamma', 'vega', 'theta', 'rho']


@export_as_api(namespace='options')
def get_greeks(order_book_ids, start_date=None, end_date=None, fields=None, model='implied_forward', market="cn"):
    """获取指定股票期权的隐含波动率iv， 以及5个希腊字母数值(delta, gamma, bega, theta, rho)
    :param order_book_ids: 股票 order_book_id or order_book_id list
    :param start_date: 开始日期, 必要参数
    :param end_date: 结束日期；默认为今天
    :param fields: str或list类型. 默认为None, 返回所有fields.
    :param model: str类型， last: 代表用每日close价格， implied_forward 代表用隐含风险收益率计算
    :param market: 默认值为"cn"
    """

    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    check_items_in_container(model, ['implied_forward', 'last'], 'model')
    if start_date is not None:
        start_date = ensure_date_int(start_date)
    else:
        raise ValueError('start_date is expected')
    if end_date is not None:
        end_date = ensure_date_int(end_date)
    else:
        end_date = ensure_date_int(datetime.datetime.now().date())
    if end_date < start_date:
        raise ValueError("invalid date range: [{!r}, {!r}]".format(start_date, end_date))

    if fields is None:
        fields = VALID_GREEKS_FIELDS
    else:
        check_items_in_container(fields, VALID_GREEKS_FIELDS, 'Greeks')

    data = get_client().execute("options.get_greeks", order_book_ids, start_date, end_date, fields, model, market=market)
    if not data:
        return None

    df = pd.DataFrame(data)
    df.set_index(["order_book_id", "trading_date"], inplace=True)
    df.sort_index(inplace=True)
    return df[fields]


SPECIAL_UNDERLYING_SYMBOL = ("510050.XSHG", "510300.XSHG", "159919.XSHE")


@export_as_api(namespace='options')
def get_contracts(
    underlying,
    option_type=None,
    maturity=None,
    strike=None,
    trading_date=None
):
    """返回符合条件的期权

    :param underlying: 标的合约, 可以填写'M'代表期货品种的字母；也可填写'M1901'这种具体 order_book_id
    :param option_type: 期权类型, 'C'代表认购期权, 'P'代表认沽期权合约, 默认返回全部
    :param maturity: 到期月份, 如'1811'代表期权18年11月到期, 默认返回全部到期月份
    :param strike: 行权价, 向左靠档, 默认返回全部行权价
    :param trading_date: 查询日期, 默认返回当前全部

    :returns
        返回order_book_id list；如果无符合条件期权则返回空list[]
    """
    underlying = ensure_string(underlying, "underlying").upper()
    instruments_df = all_instruments(type='Option')
    underlying_symbols = instruments_df.underlying_symbol.unique()
    underlying_order_book_ids = instruments_df.underlying_order_book_id.unique()
    instruments_df = all_instruments(type='Option', date=trading_date)
    if underlying in underlying_symbols:
        instruments_df = instruments_df[instruments_df.underlying_symbol == underlying]
    elif underlying in underlying_order_book_ids:
        instruments_df = instruments_df[instruments_df.underlying_order_book_id == underlying]
    else:
        raise ValueError("Unknown underlying")
    if instruments_df.empty:
        return []

    if option_type is not None:
        option_type = ensure_string(option_type, "option_type").upper()
        ensure_string_in(option_type, {'P', 'C'}, "option_type")
        instruments_df = instruments_df[instruments_df.option_type == option_type]

    if maturity is not None:
        maturity = int(maturity)
        month = maturity % 100
        if month not in range(1, 13):
            raise ValueError("Unknown month")
        year = maturity // 100 + 2000
        str_month = str(month)
        if len(str_month) == 1:
            str_month = '0' + str_month
        date_str = str(year) + '-' + str_month
        instruments_df = instruments_df[instruments_df.maturity_date.str.startswith(date_str)]
        if instruments_df.empty:
            return []

    if strike:
        if underlying in SPECIAL_UNDERLYING_SYMBOL and trading_date:
            order_book_ids = instruments_df.order_book_id.tolist()

            strikes = get_price(order_book_ids, start_date=trading_date, end_date=trading_date, fields='strike_price',
                                expect_df=is_panel_removed)
            if strikes is None:
                return []
            if is_panel_removed:
                strikes.reset_index(level=1, inplace=True, drop=True)
            else:
                strikes = strikes.T

            instruments_df.set_index(instruments_df.order_book_id, inplace=True)
            instruments_df['strike_price'] = strikes[strikes.columns[0]]
            instruments_df = instruments_df[instruments_df.strike_price.notnull()]
            if instruments_df.empty:
                return []

        l = []
        for date in instruments_df.maturity_date.unique():
            df = instruments_df[instruments_df.maturity_date == date]
            df = df[df.strike_price <= strike]
            if df.empty:
                continue
            df = df[df.strike_price.rank(method='min', ascending=False) == 1]
            l += df.order_book_id.tolist()
        return l

    return instruments_df.order_book_id.tolist()


VALID_CONTRACT_PROPERTY_FIELDS = ['product_name', 'symbol', 'contract_multiplier', 'strike_price']


def _get_multi_index(oids, end_date, listed, de_listed):
    mult = []
    tds = get_trading_dates('20150101', datetime.date.today())
    index = pd.to_datetime(tds)
    for oid in oids:
        if oid not in listed:
            continue
        s = str(listed[oid])
        e = str(min(end_date, de_listed[oid]))
        _start, _end = pd.Timestamp(s), pd.Timestamp(e)
        start_pos, end_pos = index.searchsorted(_start), index.searchsorted(_end)
        _index = index[start_pos:end_pos+1]
        mult.extend([(oid, i) for i in _index])
    return pd.MultiIndex.from_tuples(mult, names=['order_book_id', 'trading_date'])


@export_as_api(namespace='options')
def get_contract_property(
    order_book_ids,
    start_date=None,
    end_date=None,
    fields=None,
    market='cn'
):
    """获取期权合约属性
    :param order_book_ids: 股票 order_book_id or order_book_id list
    :param start_date: 开始日期, 必要参数
    :param end_date: 结束日期；默认为今天
    :param fields: str或list类型. 默认为None, 返回所有fields.
    :param market: 默认值为"cn", 可选：cn

    :returns
        返回DataFrame, 索引为 MultiIndex([order_book_id, trading_date])
    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    listed_dates = {}
    de_listed_dates = {}
    for oid in order_book_ids:
        i = instruments(oid)
        listed_dates[oid] = int(i.listed_date.replace('-', ''))
        de_listed_dates[oid] = int(i.de_listed_date.replace('-', ''))

    end_date = datetime.date.today() if not end_date else end_date
    end_date = ensure_date_int(end_date)
    if start_date:
        start_date, end_date = ensure_date_range(start_date, end_date)
        # 如果指定start_date, 将退市的order_book_id过滤掉
        order_book_ids = [oid for oid in order_book_ids if start_date <= de_listed_dates[oid]]

    if fields is None:
        fields = VALID_CONTRACT_PROPERTY_FIELDS[:]
    else:
        if not isinstance(fields, list):
            fields = [fields]
        check_items_in_container(fields, VALID_CONTRACT_PROPERTY_FIELDS, 'Contract Property')
    # get data from server
    data = get_client().execute('options.get_contract_property', order_book_ids, fields)
    if not data:
        return
    df = pd.DataFrame(data)

    index = _get_multi_index(order_book_ids, end_date, listed_dates, de_listed_dates)
    df = df.set_index(['order_book_id', 'trading_date'])
    df = df.reindex(index).ffill()
    if start_date:
        msk = df.index.get_level_values('trading_date') >= pd.Timestamp(str(start_date))
        df = df[msk]
    return df.sort_index()




# coding: utf-8
from sqlalchemy import Numeric, Column
from .fundamental_base_sql import FundamentalBase


class StkCashGen(FundamentalBase):
    cash_received_from_sales_of_goods = Column(Numeric(18, 2))
    rental_cash = Column(Numeric(18, 2))
    refunds_of_vat = Column(Numeric(18, 2))
    refunds_of_other_taxes = Column(Numeric(18, 2))
    refunds_of_taxes = Column(Numeric(18, 2))
    cash_from_other_operating_activities = Column(Numeric(18, 2))
    cash_from_operating_activities = Column(Numeric(18, 2))
    cash_paid_for_goods_and_services = Column(Numeric(18, 2))
    cash_paid_for_rental = Column(Numeric(18, 2))
    cash_paid_for_employee = Column(Numeric(18, 2))
    cash_paid_for_taxes = Column(Numeric(18, 2))
    cash_paid_for_other_operation_activities = Column(Numeric(18, 2))
    cash_paid_for_operation_activities = Column(Numeric(18, 2))
    cash_flow_from_operating_activities = Column(Numeric(18, 2))
    cash_received_from_disposal_of_investment = Column(Numeric(18, 2))
    cash_received_from_dividend = Column(Numeric(18, 2))
    cash_received_from_interest = Column(Numeric(18, 2))
    cash_received_from_disposal_of_asset = Column(Numeric(18, 2))
    cash_received_from_other_investment_activities = Column(Numeric(18, 2))
    cash_received_from_investment_activities = Column(Numeric(18, 2))
    cash_paid_for_asset = Column(Numeric(18, 2))
    cash_paid_to_acquire_investment = Column(Numeric(18, 2))
    cash_paid_for_other_investment_activities = Column(Numeric(18, 2))
    cash_paid_for_investment_activities = Column(Numeric(18, 2))
    cash_flow_from_investing_activities = Column(Numeric(18, 2))
    cash_received_from_equity_investors = Column(Numeric(18, 2))
    cash_received_from_debt_investors = Column(Numeric(18, 2))
    cash_received_from_investors = Column(Numeric(18, 2))
    cash_received_from_financial_institution_borrows = Column(Numeric(18, 2))
    cash_received_from_other_financing_activities = Column(Numeric(18, 2))
    cash_received_from_financing_activities = Column(Numeric(18, 2))
    cash_paid_for_debt = Column(Numeric(18, 2))
    cash_paid_for_dividend_and_interest = Column(Numeric(18, 2))
    cash_paid_for_other_financing_activities = Column(Numeric(18, 2))
    cash_paid_to_financing_activities = Column(Numeric(18, 2))
    cash_flow_from_financing_activities = Column(Numeric(18, 2))

    assets_depreciation_reserves = Column(Numeric(18, 2))
    fixed_asset_depreciation = Column(Numeric(18, 2))
    intangible_asset_amortization = Column(Numeric(18, 2))
    deferred_expense_amort = Column(Numeric(18, 2))

    net_inc_cash_and_equivalents = Column(Numeric(19, 4))

    exchange_rate_change_effect = Column(Numeric(19, 4))
    cash_equivalent_increase = Column(Numeric(19, 4))

    net_cash_deal_from_sub = Column(Numeric(19, 4))
    net_cash_payment_from_sub = Column(Numeric(19, 4))



# -*- coding: utf-8 -*-
from rqdatac.validators import ensure_date_or_today_int
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api


@export_as_api
def concept_list(date=None, market="cn"):
    """获取所有股票概念.

    :param date: 可指定日期，默认按当前日期返回.
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :returns: 符合指定日期内出现过的所有概念列表

    """
    date = ensure_date_or_today_int(date)
    return get_client().execute("concept_list", date, market=market)


@export_as_api
def concept(*concepts, **kwargs):
    """获取对应某个概念的股票列表。

    可指定日期，默认按当前日期返回。目前支持的概念列表可以查询以下网址:
    https://www.ricequant.com/api/research/chn#concept-API-industry

    :param concepts: 概念字符串,如 '民营医院'
    :param date: 可指定日期，默认按当前日期返回.
    :param market: 地区代码, 如 'cn'
    :returns: 符合对应概念的股票列表

    """
    date = kwargs.pop("date", None)
    market = kwargs.pop("market", "cn")
    date = ensure_date_or_today_int(date)
    if kwargs:
        raise ValueError('unknown kwargs: {}'.format(kwargs))
    return get_client().execute("concept", concepts, date, market=market)


@export_as_api
def concept_names(order_book_id, date=None, expect_type="str", market="cn"):
    """获取证券所属的概念列表。

    :param order_book_id: 证券ID
    :param date: 可指定日期，默认按当前日期返回。
    :param expect_type: 期望返回结果类型，可选址为："str"：返回字符串，"list"：返回列表，默认为str。
    :param market: 地区代码, 如 "cn" (Default value = "cn")
    :returns: 概念列表

    """

    date = ensure_date_or_today_int(date)
    data = get_client().execute("concept_names", order_book_id, date, market=market)
    if expect_type == "str":
        return data
    elif expect_type == "list":
        return data.split("|")
    raise ValueError("expect_type should be str like 'str' or 'list'")




# coding: utf-8
from sqlalchemy import Numeric, Column
from .fundamental_base_sql import FundamentalBase


class StkBalaGen(FundamentalBase):
    cash = Column(Numeric(18, 2))
    financial_asset_held_for_trading = Column(Numeric(18, 2))
    cash_equivalent = Column(Numeric(18, 2))
    current_investment = Column(Numeric(18, 2))
    current_investment_reserve = Column(Numeric(18, 2))
    net_current_investment = Column(Numeric(18, 2))
    bill_receivable = Column(Numeric(18, 2))
    devidend_receivable = Column(Numeric(18, 2))
    interest_receivable = Column(Numeric(18, 2))
    accts_receivable = Column(Numeric(18, 2))
    other_accts_receivable = Column(Numeric(18, 2))
    bad_debt_reserve = Column(Numeric(18, 2))
    net_accts_receivable = Column(Numeric(18, 2))
    other_receivables = Column(Numeric(18, 2))
    prepayment = Column(Numeric(18, 2))
    subsidy_receivable = Column(Numeric(18, 2))
    prepaid_tax = Column(Numeric(18, 2))
    inventory = Column(Numeric(18, 2))
    inventory_depreciation_reserve = Column(Numeric(18, 2))
    net_inventory = Column(Numeric(18, 2))
    deferred_expense = Column(Numeric(18, 2))
    contract_work = Column(Numeric(18, 2))
    long_term_debt_due_one_year = Column(Numeric(18, 2))
    non_current_debt_due_one_year = Column(Numeric(18, 2))
    other_current_assets = Column(Numeric(18, 2))
    current_assets = Column(Numeric(18, 2))
    financial_asset_available_for_sale = Column(Numeric(18, 2))
    financial_asset_hold_to_maturity = Column(Numeric(18, 2))
    real_estate_investment = Column(Numeric(18, 2))
    long_term_equity_investment = Column(Numeric(18, 2))
    long_term_receivables = Column(Numeric(18, 2))
    long_term_debt_investment = Column(Numeric(18, 2))
    other_long_term_investment = Column(Numeric(18, 2))
    long_term_investment = Column(Numeric(18, 2))
    provision_long_term_investment = Column(Numeric(18, 2))
    net_long_term_equity_investment = Column(Numeric(18, 2))
    net_long_term_debt_investment = Column(Numeric(18, 2))
    net_long_term_investment = Column(Numeric(18, 2))
    cost_fixed_assets = Column(Numeric(18, 2))
    accumulated_depreciation = Column(Numeric(18, 2))
    net_val_fixed_assets = Column(Numeric(18, 2))
    depreciation_reserve = Column(Numeric(18, 2))
    net_fixed_assets = Column(Numeric(18, 2))
    engineer_material = Column(Numeric(18, 2))
    construction_in_progress = Column(Numeric(18, 2))
    fixed_asset_to_be_disposed = Column(Numeric(18, 2))
    capitalized_biological_assets = Column(Numeric(18, 2))
    oil_and_gas_assets = Column(Numeric(18, 2))
    total_fixed_assets = Column(Numeric(18, 2))
    intangible_assets = Column(Numeric(18, 2))
    impairment_intangible_assets = Column(Numeric(18, 2))
    goodwill = Column(Numeric(18, 2))
    deferred_charges = Column(Numeric(18, 2))
    long_term_deferred_expenses = Column(Numeric(18, 2))
    other_long_term_assets = Column(Numeric(18, 2))
    total_intangible_and_other_assets = Column(Numeric(18, 2))
    deferred_income_tax_assets = Column(Numeric(18, 2))
    other_non_current_assets = Column(Numeric(18, 2))
    non_current_assets = Column(Numeric(18, 2))
    total_assets = Column(Numeric(18, 2))
    short_term_loans = Column(Numeric(18, 2))
    financial_liabilities = Column(Numeric(18, 2))
    notes_payable = Column(Numeric(18, 2))
    accts_payable = Column(Numeric(18, 2))
    advance_from_customers = Column(Numeric(18, 2))
    proxy_sale_revenue = Column(Numeric(18, 2))
    payroll_payable = Column(Numeric(18, 2))
    walfare_payable = Column(Numeric(18, 2))
    dividend_payable = Column(Numeric(18, 2))
    tax_payable = Column(Numeric(18, 2))
    interest_payable = Column(Numeric(18, 2))
    other_fees_payable = Column(Numeric(18, 2))
    internal_accts_payable = Column(Numeric(18, 2))
    other_payable = Column(Numeric(18, 2))
    short_term_debt = Column(Numeric(18, 2))
    accrued_expense = Column(Numeric(18, 2))
    estimated_liabilities = Column(Numeric(18, 2))
    deferred_income = Column(Numeric(18, 2))
    long_term_liabilities_due_one_year = Column(Numeric(18, 2))
    other_current_liabilities = Column(Numeric(18, 2))
    current_liabilities = Column(Numeric(18, 2))
    long_term_loans = Column(Numeric(18, 2))
    bond_payable = Column(Numeric(18, 2))
    long_term_payable = Column(Numeric(18, 2))
    grants_received = Column(Numeric(18, 2))
    housing_revolving_funds = Column(Numeric(18, 2))
    other_long_term_liabilities = Column(Numeric(18, 2))
    long_term_liabilities = Column(Numeric(18, 2))
    deferred_income_tax_liabilities = Column(Numeric(18, 2))
    other_non_current_liabilities = Column(Numeric(18, 2))
    non_current_liabilities = Column(Numeric(18, 2))
    total_liabilities = Column(Numeric(18, 2))
    paid_in_capital = Column(Numeric(18, 2))
    invesment_refund = Column(Numeric(18, 2))
    capital_reserve = Column(Numeric(18, 2))
    surplus_reserve = Column(Numeric(18, 2))
    statutory_reserve = Column(Numeric(18, 2))
    unrealised_investment_loss = Column(Numeric(18, 2))
    undistributed_profit = Column(Numeric(18, 2))
    equity_parent_company = Column(Numeric(18, 2))
    total_equity = Column(Numeric(18, 2))
    minority_interest = Column(Numeric(18, 2))
    total_equity_and_liabilities = Column(Numeric(18, 2))
    provision = Column(Numeric(18, 2))
    deferred_revenue = Column(Numeric(18, 2))

    non_current_liability_due_one_year = Column(Numeric(18, 2))
    liability_prefer_stock = Column(Numeric(18, 2))
    equity_prefer_stock = Column(Numeric(18, 2))
    equity_preferred_stock = Column(Numeric(18, 2))

    unrealised_investment_losses = Column(Numeric(19, 4))
    income_tax_refund = Column(Numeric(19, 4))
    non_current_asset_due_one_year = Column(Numeric(19, 4))
    long_term_deferred_income = Column(Numeric(19, 4))
    dividend_receivable = Column(Numeric(19, 4))



# -*- coding: utf-8 -*-
import datetime
import pandas as pd
from rqdatac.validators import (
    ensure_string,
    ensure_list_of_string,
    check_items_in_container,
    ensure_date_range,
    ensure_date_int,
    ensure_order_book_ids,
)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api

FIELDS = (
    "new_comments",
    "total_comments",
    "new_followers",
    "total_followers",
    "sell_actions",
    "buy_actions",
)


@export_as_api(namespace="xueqiu")
def top_stocks(field, date, frequency="1d", count=5, market="cn"):
    """获取雪球舆情数据

    :param field: 如 'new_comments', 'total_comments', 'new_followers',
                 'total_followers', 'sell_actions', 'buy_actions'
    :param date: 如 '2015-05-21', 必须在2015年4月23日之后
    :param frequency: 如 '1d', '1w', '1M' (Default value = "1d")
    :param count: 如 5, 10, 100 (Default value = 5)
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :returns: 如果有数据，返回一个DataFrame,否则返回None

    """
    field = ensure_string(field, "field")
    frequency = ensure_string(frequency, "frequency")
    check_items_in_container([field], FIELDS, "field")
    check_items_in_container([frequency], {"1d", "1w", "1M"}, "frequency")
    d = {"1d": "d", "1M": "m", "1w": "w"}
    frequency = d[frequency]
    date = ensure_date_int(date)
    if date < 20150423 or date > ensure_date_int(datetime.datetime.today()):
        raise ValueError("date out of range, start_date " "cannot be earlier than 2015-04-23")
    data = get_client().execute("xueqiu.top_stocks", field, date, frequency, count, market)

    if not data:
        return
    df = pd.DataFrame(data)
    df = df[["order_book_id", field]]
    return df


@export_as_api(namespace="xueqiu")
def history(
    order_book_ids,
    start_date="2015-05-21",
    end_date="2016-05-21",
    frequency="1d",
    fields=None,
    market="cn",
):
    """获取雪球历史舆情数据

    :param order_book_ids: 股票代码或代码列表
    :param start_date: 如 '2015-05-21', 必须在2015年4月23日之后 (Default value = "2015-05-21")
    :param end_date: 如 '2016-05-21' (Default value = "2016-05-21")
    :param frequency: 如 '1d' (Default value = "1d")
    :param fields: 如 'new_comments', 'total_comments', 'new_followers',
                 'total_followers', 'sell_actions', 'buy_actions' (Default value = None)
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :returns: 返回pd.Panel或pd.DataFrame或pd.Series

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    if fields:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, FIELDS, "fields")
    else:
        fields = FIELDS
    frequency = ensure_string(frequency, "frequency")
    check_items_in_container([frequency], {"1d"}, "frequency")
    start_date, end_date = ensure_date_range(start_date, end_date)

    if start_date < 20150423:
        raise ValueError("date out of range, start_date " "cannot be earlier than 2015-04-23")
    data = get_client().execute(
        "xueqiu.history", order_book_ids, start_date, end_date, fields, market
    )
    if not data:
        return
    df = pd.DataFrame(data)
    df = df.set_index(["date", "order_book_id"])
    df.sort_index(inplace=True)
    pl = df.to_panel()
    if len(pl.minor_axis) == 1:
        pl = pl.minor_xs(pl.minor_axis[0])
    if len(fields) == 1:
        pl = pl[fields[0]]
    return pl



# -*- coding: utf-8 -*-
import os
import time
import ssl

from rqdatac.validators import ensure_list_of_string
from rqdatac.decorators import export_as_api, retry

try:
    from orjson import dumps as json_dumps, loads as json_loads
except ImportError:
    try:
        import rapidjson as json
    except ImportError:
        import json
    def json_dumps(*args, **kwargs):
        return json.dumps(*args, **kwargs).encode('utf-8')

    def json_loads(*args, **kwargs):
        return json.loads(*args, **kwargs)

@export_as_api
class LiveMarketDataClient():
    def __init__(self, ws_server_uri="wss://rqdata.ricequant.com/live_md"):
        """websocket对象初始化

        :param ws_server_uri: websocket服务地址, 如 wss://rqdata.ricequant.com/live_md

        """
        self._info = None
        self._client = None
        self._ws_server_uri = ws_server_uri
        self._subscribed = set()
        self._init_websocket_client(self._ws_server_uri)

    def _init_websocket_client(self, ws_server_uri):
        from rqdatac.client import get_client
        _token = get_client().execute(
            "user.get_live_md_auth_token",
        )

        try:
            import websocket
        except ImportError:
            raise ImportError(
                "LiveMarketDataClient requires websocket-client package; run 'pip install websocket-client' to fix.")

        login_data = {
            "action": "auth_by_token",
            "token": _token
        }
        _websocket_client = websocket.WebSocket()
        self._client = _websocket_client
        retry(suppress_exceptions=(websocket.WebSocketException,), count=3)(self._websocket_login)(login_data)

    def _websocket_login(self, login_data):
        try:
            self._client.connect(self._ws_server_uri)
        except ssl.SSLCertVerificationError:
            self._client.sock_opt.sslopt = {"cert_reqs": ssl.CERT_NONE}
            self._client.connect(self._ws_server_uri)
        self._client.send(json_dumps(login_data))
        res = self._client.recv()
        self._info = json_loads(res)


    @property
    def info(self):
        return self._info

    def subscribe(self, channels):
        """订阅实时行情

        :param channels: 订阅的标的列表 分钟和tick分别以 bar_ 和tick_开头 以平安银行为例，
            subscribe('bar_000001.XSHE')  # 订阅分钟线的实时行情
            subscribe('tick_000001.XSHE')  # 订阅tick的实时行情
            可以同时订阅多支标的 subscribe(['bar_000001.XSHE'， 'bar_000002.XSHE')

        """
        channels = ensure_list_of_string(channels)
        data = {
            "action": "subscribe",
            "channels": channels,
        }
        self._client.send(json_dumps(data))

    def unsubscribe(self, channels):
        """取消订阅实时行情

        :param channels: 取消订阅的标的列表 分钟和tick分别以 bar_ 和tick_开头 以平安银行为例，
            unsubscribe('bar_000001.XSHE')  # 订阅分钟线的实时行情
            unsubscribe('tick_000001.XSHE')  # 订阅tick的实时行情

        """
        channels = ensure_list_of_string(channels)
        data = {
            "action": "unsubscribe",
            "channels": channels,
        }
        self._client.send(json_dumps(data))

    def _reconnect(self):
        self._init_websocket_client(self._ws_server_uri)
        self.subscribe(list(self._subscribed))

    def listen(self):
        """获取订阅实时行情

        :returns: 实时行情信息

        """
        from websocket import WebSocketException
        while True:
            try:
                res = self._client.recv()
                if res:
                    data = json_loads(res)
                    if data['action'] == 'feed':
                        yield data
                    elif data['action'] == 'subscribe_reply':
                        self._subscribed.update(data['subscribed'])
                    elif data['action'] == 'unsubscribe_reply':
                        self._subscribed -= set(data['unsubscribed'])
                time.sleep(0.1)
            except WebSocketException:
                self._reconnect()



# -*- coding: utf-8 -*-
from decimal import Decimal, DecimalTuple


import pandas as pd

from rqdatac.client import get_client
from rqdatac.decorators import export_as_api
from rqdatac.validators import ensure_list_of_string

_SIGN = "sign"
_DIGITS = "digits"
_EXPONENT = "exponent"


def _encode_decimal(num, prefix):
    decimal_tuple = Decimal(str(num)).normalize().as_tuple()
    prefix += "."
    encoded_decimal = {
        prefix + _SIGN: decimal_tuple.sign,
        prefix + _DIGITS: decimal_tuple.digits,
        prefix + _EXPONENT: decimal_tuple.exponent,
    }
    return encoded_decimal


def _decode_decimal(encoded_decimal):
    sign = encoded_decimal[_SIGN]
    digits = tuple(encoded_decimal[_DIGITS])
    exponent = encoded_decimal[_EXPONENT]
    return Decimal(DecimalTuple(sign=sign, digits=digits, exponent=exponent))


def _remake_decimal(dlist):
    for one in dlist:
        for key, value in one.items():
            if isinstance(value, dict) and value.get(_DIGITS):
                one[key] = _decode_decimal(value)


@export_as_api(namespace="fenji")
def get_a_by_yield(current_yield, listing=True, market="cn"):
    if listing is not None:
        listing = bool(listing)

    data = get_client().execute("fenji.get_a_by_yield", current_yield, listing, market)
    return data


@export_as_api(namespace="fenji")
def get_a_by_interest_rule(interest_rule, listing=True, market="cn"):
    if listing is not None:
        listing = bool(listing)
    data = get_client().execute("fenji.get_a_by_interest_rule", interest_rule, listing, market)
    return data


@export_as_api(namespace="fenji")
def get_all(field_list=None, market="cn"):
    data = get_client().execute("fenji.get_all", field_list, market)
    _remake_decimal(data)
    df = pd.DataFrame(data)
    df.reindex(columns=sorted(df.columns))
    return df


@export_as_api(namespace="fenji")
def get(order_book_ids, field_list=None, market="cn"):
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    data = get_client().execute("fenji.get", order_book_ids, field_list, market)
    if not data:
        return
    _remake_decimal(data)
    df = pd.DataFrame(data)
    df.reindex(columns=sorted(df.columns))
    return df



# -*- coding: utf-8 -*-


class RQDataError(Exception):
    rqerrno = 1
    DEFAULT_MSG = "default rqdata error"

    def __str__(self):
        s = super(Exception, self).__str__()
        if s:
            return s
        return self.DEFAULT_MSG

    def __repr__(self):
        return "{}: {}".format(self.__class__.__name__, self.__str__())


class AuthenticationFailed(RQDataError):
    rqerrno = 2
    DEFAULT_MSG = "authentication failed."


class ErrorFromBackend(RQDataError):
    rqerrno = 3
    DEFAULT_MSG = "raise error from backend"
    pass


class PermissionDenied(ErrorFromBackend):
    rqerrno = 4
    DEFAULT_MSG = "permission denied"


class NoSuchService(ErrorFromBackend):
    rqerrno = 7
    DEFAULT_MSG = "Can't found the request service"


class QuotaExceeded(ErrorFromBackend):
    rqerrno = 5
    DEFAULT_MSG = "quota exceeded"


class MarketNotSupportError(ErrorFromBackend):
    rqerrno = 6
    DEFAULT_MSG = "market not supported yet."


class InternalError(RQDataError):
    rqerrno = -1
    DEFAULT_MSG = "Server raised an error and can't handle it."


class GatewayError(InternalError):
    rqerrno = -2
    DEFAULT_MSG = "Can't communicate with gateway."


class BadRequest(ErrorFromBackend):
    rqerrno = 400
    DEFAULT_MSG = "Bad Request Content"


class OverwriteWarning(Warning):
    pass


_ERROR_MAP = {}


def __go(lcs):
    global _ERROR_MAP
    _ERROR_MAP = {v.rqerrno: v for k, v in lcs.items() if hasattr(v, "rqerrno")}


__go(locals())
del __go


def get_error(errorno):
    return _ERROR_MAP.get(errorno, RQDataError)


__all__ = tuple(i.__name__ for i in _ERROR_MAP.values()) + ("get_error",)



# -*- coding: utf-8 -*-
import datetime
import warnings

import pandas as pd

from rqdatac.utils import to_datetime, is_panel_removed, to_date
from rqdatac.validators import (
    ensure_date_range,
    ensure_date_or_today_int,
    ensure_list_of_string,
    check_items_in_container,
    ensure_order,
    ensure_order_book_id,
    ensure_order_book_ids,
    ensure_dates_base_on_listed_date,
    ensure_string,
    ensure_date_int,
    raise_for_no_panel,
)
from rqdatac.services.basic import instruments
from rqdatac.services.calendar import (
    get_trading_dates,
    get_previous_trading_date,
    get_trading_dates_in_type,
)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api, compatible_with_parm


@export_as_api
def current_stock_connect_quota(
        connect=None, fields=None
):
    """
    获取沪港通、深港通资金流向数据

    :param connect: 字符串，目前可从sh_to_hk, hk_to_sh, sz_to_hk, hk_to_sz多选, 默认为查询所有资金流向数据
    :param fields:  字符串列表，需要的字段，目前可从buy_turnover, sell_turnover, quota_balance, quota_balance_ratio多选, 默认取所有

    """
    DEFAULT_CONNECT = ["sh_to_hk", "hk_to_sh", "sz_to_hk", "hk_to_sz"]
    if connect is None:
        connect = DEFAULT_CONNECT
    else:
        connect = ensure_list_of_string(connect)
        check_items_in_container(connect, DEFAULT_CONNECT, 'connect')

    DEFAULT_FIELDS = ['buy_turnover', 'sell_turnover', 'quota_balance', 'quota_balance_ratio']
    if fields is None:
        fields = DEFAULT_FIELDS
    else:
        fields = ensure_list_of_string(fields)
        check_items_in_container(fields, DEFAULT_FIELDS, 'fields')

    data = get_client().execute("current_stock_connect_quota", connect=connect)
    res = pd.DataFrame(data)
    if res.empty:
        return None
    res["datetime"] = pd.to_datetime(res["datetime"], format='%Y%m%d%H%M')
    res.set_index(['datetime', 'connect'], inplace=True)
    res = res[fields]
    return res


@export_as_api
def get_stock_connect_quota(connect=None, start_date=None, end_date=None, fields=None):
    """获取历史沪深港通额度日频数据

    :param connect: 默认返回全部content ["sh_to_hk", "hk_to_sh", "sz_to_hk", "hk_to_sz"]
    :param start_date: 默认为全部历史数据
    :param end_date: 默认为最新日期
    :param fields:  默认为所有字段 ['buy_turnover', 'sell_turnover', 'quota_balance', 'quota_balance_ratio']
    :return:
    """
    DEFAULT_CONNECT = ["sh_to_hk", "hk_to_sh", "sz_to_hk", "hk_to_sz"]
    if connect is None:
        connect = DEFAULT_CONNECT
    else:
        connect = ensure_list_of_string(connect)
        check_items_in_container(connect, DEFAULT_CONNECT, 'connect')

    DEFAULT_FIELDS = ['buy_turnover', 'sell_turnover', 'quota_balance', 'quota_balance_ratio']
    if fields is None:
        fields = DEFAULT_FIELDS
    else:
        fields = ensure_list_of_string(fields)
        check_items_in_container(fields, DEFAULT_FIELDS, 'fields')

    start_date = ensure_date_int(start_date) if start_date else start_date
    end_date = ensure_date_int(end_date) if end_date else end_date

    if start_date and end_date and start_date > end_date:
        raise ValueError("invalid date range: [{!r}, {!r}]".format(start_date, end_date))

    data = get_client().execute(
        "get_stock_connect_quota", connect=connect, start_date=start_date, end_date=end_date, fields=fields
    )
    if not data:
        return None
    res = pd.DataFrame(data)
    res.set_index(['datetime', 'connect'], inplace=True)
    res.sort_index(ascending=True, inplace=True)
    return res


@export_as_api
def is_st_stock(order_book_ids, start_date=None, end_date=None, market="cn"):
    """判断股票在给定的时间段是否是ST股, 返回值为一个DataFrame

    :param order_book_ids: 股票 id
    :param start_date:  (Default value = None)
    :param end_date:  (Default value = None)
    :param market:  (Default value = "cn")

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, type="CS", market=market)

    if len(order_book_ids) == 1:
        instrument = instruments(order_book_ids[0], market=market)
        start_date, end_date = ensure_dates_base_on_listed_date(instrument, start_date, end_date, market)
        if start_date is None:
            return

    start_date, end_date = ensure_date_range(start_date, end_date)

    trading_dates = pd.to_datetime(get_trading_dates(start_date, end_date, market=market))
    data = get_client().execute(
        "get_st_days", order_book_ids, start_date=start_date, end_date=end_date
    )
    df = pd.DataFrame(data=False, columns=order_book_ids, index=trading_dates)
    for idx, dates in data.items():
        for date in dates:
            date = to_datetime(date)
            df.at[date, idx] = True
    return df


@export_as_api
def _is_st_stock(order_book_id, date=None, market="cn"):
    """判断股票在给定日期是否是ST股
    :param order_book_id: 股票id
    :param date:  (Default value = None)
    :param market:  (Default value = "cn")
    :returns: True or False
    """
    order_book_id = ensure_order_book_id(order_book_id, type="CS", market=market)
    date = ensure_date_or_today_int(date)
    df = is_st_stock(order_book_id, start_date=date, end_date=date, market=market)
    if df is None or df.empty:
        return False
    else:
        return df[order_book_id][0]


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def is_suspended(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取股票停牌信息

    :param order_book_ids: 股票名称
    :param start_date: 开始日期, 如'2013-01-04' (Default value = None)
    :param end_date: 结束日期，如'2014-01-04' (Default value = None)
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :returns: DataFrame

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, type="CS", market=market)

    if len(order_book_ids) == 1:
        instrument = instruments(order_book_ids[0], market=market)
        start_date, end_date = ensure_dates_base_on_listed_date(instrument, start_date, end_date, market)
        if start_date is None:
            return
    if end_date is None:
        end_date = datetime.date.today()
    start_date, end_date = ensure_date_range(start_date, end_date)

    trading_dates = pd.to_datetime(get_trading_dates(start_date, end_date, market=market))
    df = pd.DataFrame(data=False, columns=order_book_ids, index=trading_dates)
    data = get_client().execute("get_suspended_days", order_book_ids, start_date, end_date, market=market)
    for idx, dates in data.items():
        for date in dates:
            date = to_datetime(int(date))
            df.at[date, idx] = True
    df.sort_index(inplace=True)
    return df


stock_fields = {"shares_holding": "shares_holding", "holding_ratio": "holding_ratio"}
special_symbols = ["all_connect", "shanghai_connect", "shenzhen_connect"]
symbols_map = {"shanghai_connect": "SH", "shenzhen_connect": "SZ"}


@export_as_api
def get_stock_connect(order_book_ids, start_date=None, end_date=None, fields=None, expect_df=True):
    """获取"陆股通"的持股、持股比例

    :param order_book_ids: 股票列表
    :param start_date: 开始日期: 如'2017-03-17' (Default value = None)
    :param end_date: 结束日期: 如'2018-03-16' (Default value = None)
    :param fields: 默认为所有字段，可输入shares_holding, holding_ratio, adjusted_holding_ratio (Default value = None)
    :param expect_df: 是否返回 MultiIndex DataFrame (Default value = True)
    :returns: 返回pandas.DataFrame or pandas.Panel

    """
    if order_book_ids not in ("shanghai_connect", "shenzhen_connect", "all_connect"):
        order_book_ids = ensure_order_book_ids(order_book_ids, type="CS")
    start_date, end_date = ensure_date_range(start_date, end_date)
    if fields is not None:
        fields = ensure_list_of_string(fields)
        for f in fields:
            if f not in ("shares_holding", "holding_ratio", "adjusted_holding_ratio"):
                raise ValueError("invalid field: {}".format(f))
    else:
        fields = ["shares_holding", "holding_ratio", "adjusted_holding_ratio"]
    data = get_client().execute("get_stock_connect", order_book_ids, start_date, end_date, fields)
    if not data:
        return None
    df = pd.DataFrame(data, columns=["trading_date", "order_book_id"] + fields)

    if not expect_df and not is_panel_removed:
        df = df.set_index(["trading_date", "order_book_id"])
        df = df.to_panel()
        df.major_axis.name = None
        df.minor_axis.name = None
        if len(order_book_ids) == 1:
            df = df.minor_xs(order_book_ids[0])
        if len(fields) == 1:
            df = df[fields[0]]
        if len(order_book_ids) != 1 and len(fields) != 1:
            warnings.warn("Panel is removed after pandas version 0.25.0."
                          " the default value of 'expect_df' will change to True in the future.")
        return df
    else:
        df.sort_values(["order_book_id", "trading_date"], inplace=True)
        df.set_index(["order_book_id", "trading_date"], inplace=True)
        if expect_df:
            return df

        if len(order_book_ids) != 1 and len(fields) != 1:
            raise_for_no_panel()

        if len(order_book_ids) == 1:
            df.reset_index(level=0, drop=True, inplace=True)
            if len(fields) == 1:
                df = df[fields[0]]
            return df
        else:
            df = df.unstack(0)[fields[0]]
            df.index.name = None
            df.columns.name = None
            return df


MARGIN_FIELDS = (
    "margin_balance",
    "buy_on_margin_value",
    "short_sell_quantity",
    "margin_repayment",
    "short_balance_quantity",
    "short_repayment_quantity",
    "short_balance",
    "total_balance",
)

MARGIN_SUMMARY_MAP = {"SH": "XSHG", "XSHG": "XSHG", "SZ": "XSHE", "XSHE": "XSHE"}


@export_as_api
def get_securities_margin(
        order_book_ids, start_date=None, end_date=None, fields=None, expect_df=True, market="cn"
):
    """获取股票融资融券数据

    :param order_book_ids: 股票代码或代码列表
    :param start_date: 开始时间，支持 str, date, datetime, pandasTimestamp
        默认为 end_date 之前一个月 (Default value = None)
    :param end_date: 结束时间 默认为当前日期前一天 (Default value = None)
    :param fields: str 或 list 类型. 默认为 None, 返回所有字段。可选字段包括：
                   margin_balance, buy_on_margin_value, margin_repayment, short_balance, short_balance_quantity,
                   short_sell_quantity, short_repayment_quantity, total_balance
                   (Default value = None)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market: 地区代码, 如: 'cn' (Default value = "cn")
    :returns: 如果传入多个股票代码，且 fields 为多个或者 None，返回 pandas.Panel
        如果传入一只股票或者 fields 为单个字段，则返回 pandas.DataFrame
        如果传入的股票代码和字段数都是1，则返回 pandas.Series

    """

    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    all_list = []
    for order_book_id in order_book_ids:
        if order_book_id.upper() in MARGIN_SUMMARY_MAP:
            all_list.append(MARGIN_SUMMARY_MAP[order_book_id.upper()])
        else:
            inst = instruments(order_book_id, market)

            if inst.type in ["CS", "ETF", "LOF"]:
                all_list.append(inst.order_book_id)
            else:
                warnings.warn("{} is not stock, ETF, or LOF.".format(order_book_id))
    order_book_ids = all_list
    if not order_book_ids:
        raise ValueError("no valid securities in {}".format(order_book_ids))
    if fields is None:
        fields = list(MARGIN_FIELDS)
    else:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, MARGIN_FIELDS, "fields")
        fields = ensure_order(fields, MARGIN_FIELDS)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if end_date > ensure_date_or_today_int(None):
        end_date = ensure_date_or_today_int(get_previous_trading_date(datetime.date.today()))
    trading_dates = pd.to_datetime(get_trading_dates(start_date, end_date, market=market))

    data = get_client().execute(
        "get_securities_margin", order_book_ids, start_date, end_date, market=market
    )
    if not data:
        return

    if not expect_df and not is_panel_removed:

        pl = pd.Panel(items=fields, major_axis=trading_dates, minor_axis=order_book_ids)
        for r in data:
            for field in fields:
                value = r.get(field)
                pl.at[field, r["date"], r["order_book_id"]] = value

        if len(order_book_ids) == 1:
            pl = pl.minor_xs(order_book_ids[0])
        if len(fields) == 1:
            pl = pl[fields[0]]
        if len(order_book_ids) != 1 and len(fields) != 1:
            warnings.warn("Panel is removed after pandas version 0.25.0."
                          " the default value of 'expect_df' will change to True in the future.")
        return pl
    else:
        df = pd.DataFrame(data)
        df.sort_values(["order_book_id", "date"], inplace=True)
        df.set_index(["order_book_id", "date"], inplace=True)
        df = df.reindex(columns=fields)
        if expect_df:
            return df

        if len(order_book_ids) != 1 and len(fields) != 1:
            raise_for_no_panel()

        if len(order_book_ids) == 1:
            df.reset_index(level=0, drop=True, inplace=True)
            if len(fields) == 1:
                df = df[fields[0]]
            return df
        else:
            df = df.unstack(0)[fields[0]]
            df.index.name = None
            df.columns.name = None
            return df


MARGIN_TYPE = ("stock", "cash")
EXCHANGE_TYPE = {"SZ": "XSHE", "sz": "XSHE", "xshe": "XSHE", "SH": "XSHG", "sh": "XSHG", "xshg": "XSHG"}
EXCHANGE_CONTENT = ["XSHE", "XSHG"]


@export_as_api
def get_margin_stocks(date=None, exchange=None, margin_type='stock', market="cn"):
    """获取融资融券信息

    :param date: 查询日期，默认返回今天上一交易日，支持 str, timestamp, datetime 类型
    :param exchange: 交易所信息，默认不填写则返回全部。
                    str类型，默认为 None，返回所有字段。可选字段包括：
                    'XSHE', 'sz' 代表深交所；'XSHG', 'sh' 代表上交所，不区分大小写
                    (Default value = None)
    :param margin_type: 'stock' 代表融券卖出，'cash'，代表融资买入，默认为'stock'

    """
    if date:
        date = ensure_date_int(date)
    else:
        date = get_previous_trading_date(datetime.date.today())
        date = date.year * 10000 + date.month * 100 + date.day

    if exchange is None:
        exchange = EXCHANGE_CONTENT
    else:
        exchange = ensure_string(exchange, "exchange")
        if exchange in EXCHANGE_TYPE:
            exchange = EXCHANGE_TYPE[exchange]
        check_items_in_container(exchange, EXCHANGE_CONTENT, "exchange")
        exchange = [exchange]

    margin_type = ensure_string(margin_type, "margin_type")
    check_items_in_container(margin_type, MARGIN_TYPE, "margin_type")

    data = get_client().execute(
        "get_margin_stocks", date, exchange, margin_type, market=market
    )

    if not data:
        return []
    else:
        return sorted(data)


share_fields = {
    "total": "total_shares",
    "circulation_a": "a_cir_shares",
    "non_circulation_a": "a_non_cir_shares",
    "total_a": "a_total_shares",
}

anti_fields = {v: k for k, v in share_fields.items()}


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_shares(order_book_ids, start_date=None, end_date=None, fields=None, expect_df=True, market="cn"):
    """获取流通股本信息

    :param order_book_ids: 股票名称
    :param start_date: 开始日期, 如'2013-01-04' (Default value = None)
    :param end_date: 结束日期，如'2014-01-04' (Default value = None)
    :param fields: 如'total', 'circulation_a' (Default value = None)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market: 地区代码，如'cn' (Default value = "cn")
    :returns: 返回一个DataFrame

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if fields:
        fields = ensure_list_of_string(fields, "fields")
        if 'management_circulation' in fields:
            fields.remove('management_circulation')
            if fields:
                warnings.warn("management_circulation is removed")
            else:
                raise ValueError("management_circulation is removed")
        check_items_in_container(fields, set(share_fields), "fields")
        fields = [share_fields[i] for i in fields]
    else:
        fields = list(share_fields.values())

    all_shares = get_client().execute("get_shares", order_book_ids, fields, market=market)
    if not all_shares:
        return
    dates = get_trading_dates_in_type(start_date, end_date, expect_type="datetime", market=market)
    df = pd.DataFrame(all_shares)
    unique = set(df.order_book_id)
    for order_book_id in order_book_ids:
        if order_book_id not in unique:
            df = df.append(
                {"order_book_id": order_book_id, "date": df.date.iloc[-1]}, ignore_index=True
            )
    df.set_index(["date", "order_book_id"], inplace=True)
    df.sort_index(inplace=True)
    df = df.unstack(level=1)
    index = df.index.union(dates)
    df = df.reindex(index)
    df = df.fillna(method="ffill")
    df = df.loc[list(dates)]
    df = df.dropna(how="all")
    df = df[fields]
    if not is_panel_removed and not expect_df:
        pl = df.stack(1).to_panel()
        pl.items = [anti_fields[i] for i in pl.items]
        if len(order_book_ids) == 1:
            pl = pl.minor_xs(order_book_ids[0])
        if len(fields) == 1:
            pl = pl[anti_fields[fields[0]]]
        if len(order_book_ids) != 1 and len(fields) != 1:
            warnings.warn("Panel is removed after pandas version 0.25.0."
                          " the default value of 'expect_df' will change to True in the future.")
        return pl
    else:
        df = df.stack(1)
        df.index.set_names(["date", "order_book_id"], inplace=True)
        df = df.reorder_levels(["order_book_id", "date"]).sort_index()
        df = df.rename(columns=anti_fields)
        if expect_df:
            return df

        if len(order_book_ids) != 1 and len(fields) != 1:
            raise_for_no_panel()

        if len(order_book_ids) == 1:
            df.reset_index(level=0, drop=True, inplace=True)
            if len(fields) == 1:
                df = df[anti_fields[fields[0]]]
            return df
        else:
            df = df.unstack(0)[anti_fields[fields[0]]]
            df.index.name = None
            df.columns.name = None
            return df


allotment_fields = [
    "proportion",
    "allotted_proportion",
    "allotted_shares",
    "allotment_price",
    "book_closure_date",
    "ex_right_date", ]


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_allotment(order_book_ids, start_date=None, end_date=None, fields=None, market="cn"):
    """获取配股信息
    :param order_book_ids: 股票名称
    :param start_date: 开始日期, 如'1991-01-01' (Default value = 1991-01-01)
    :param end_date: 结束日期，如'2014-01-04' (Default value = None)
    :param market: 地区代码，如'cn' (Default value = "cn")
    :returns: 返回一个DataFrame
    'order_book_id': 股票合约代码
    'declaration_announcement_date': 首次信息发布日期
    'proportion': 配股比例(每一股对应的配股比例)
    'allotted_proportion': 实际配股比例(每一股对应的配股比例)
    'allotted_shares': 实际配股数量
    'allotment_price': 每股配股价格
    'book_closure_date': 股权登记日
    'ex_right_date': 除权除息日
    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)

    if start_date:
        start_date = ensure_date_int(start_date)
    if end_date:
        end_date = ensure_date_int(end_date)

    if fields:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, allotment_fields, "fields")
    else:
        fields = allotment_fields

    all_allotment = get_client().execute("get_allotment", order_book_ids, fields, start_date, end_date, market=market)
    if not all_allotment:
        return
    df = pd.DataFrame(all_allotment)
    df.set_index(["order_book_id", "declaration_announcement_date"], inplace=True)
    df.sort_index(inplace=True)
    df = df[fields]
    return df



# -*- coding: utf-8 -*-
import os
import sys
import versioneer
from setuptools import setup, find_packages


with open('README.md') as f:
    readme = f.read()

ext_modules = []

if os.environ.get("BUILD_EXT"):
    from Cython.Build import cythonize
    ext_modules = cythonize('rqdatac/connection.py')


requirements = [
    'brotli',
    'python-dateutil',
    'msgpack >= 0.5.2',
    'six',
    'SQLAlchemy<=1.3.24',
    'pymysql',
]

if sys.version_info < (3, 0):
    requirements.extend([
        'numpy<1.17.0',
        'pandas<0.25.0',
        'typing',
    ])
    if os.name == 'nt':
        requirements.append('win_inet_pton')
elif sys.version_info < (3, 5):
    requirements.extend([
        'numpy<1.17.0',
        'pandas<=0.22.0',
        'typing',
    ])
else:
    requirements.extend([
        'pandas',
        # 有benchmark表示这个包比orjson慢一点
        'python-rapidjson',
        'orjson'
    ])

setup(
    name="rqdatac",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Ricequant Data SDK",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Ricequant",
    author_email="public@ricequant.com",
    url="https://www.ricequant.com/doc/rqdata-institutional#research-version",
    download_url="https://pypi.org/",
    include_package_data=True,
    packages=find_packages(include=["rqdatac", "rqdatac.*"]),
    install_requires=requirements,
    extras_require={
        'subscribe': ['websocket-client'],
        'proxy': ["PySocks"],
        'fund': ["rqdatac_fund"],
        'bond': ["rqdatac_bond"],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Environment :: Console",
    ],
    zip_safe=False,
    package_data={"": ["*.*"]},
    ext_modules=ext_modules
)



# -*- coding: utf-8 -*-
import warnings

warnings.filterwarnings("ignore", message="[.\n]*Pandas[.\n]*")
warnings.simplefilter(action="ignore", category=FutureWarning)
from .client import init, reset, initialized  # noqa


__all__ = ["__version__", "init", "reset", "initialized"]


def __go():
    import importlib
    import pkgutil

    for loader, module_name, is_pkg in pkgutil.walk_packages(__path__, "rqdatac."):
        if module_name.startswith("rqdatac.services") and not is_pkg:
            importlib.import_module(module_name)


__go()

del __go

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
#
# Copyright 2018 Ricequant, Inc
import time
from threading import Lock

import numpy as np
import pandas as pd

from rqdatac.client import get_client

_lock = Lock()
_expired = time.time()
_ex_factor_cache = {}
_split_factor_cache = {}

MIN_DATE = pd.Timestamp(1800, 1, 1)


def get_ex_factor_for(order_book_ids, market):
    with _lock:
        _prepare_ex_factor_for(order_book_ids, market)
        return {ob: _ex_factor_cache[ob] for ob in order_book_ids}


def _prepare_ex_factor_for(order_book_ids, market):
    global _ex_factor_cache, _expired, _split_factor_cache
    if _expired < time.time():
        _ex_factor_cache.clear()
        _split_factor_cache.clear()
        _expired = time.time() + 3600 * 8

    missing = [ob for ob in order_book_ids if ob not in _ex_factor_cache]
    if not missing:
        return

    data = get_client().execute('_get_ex_factor', missing, market)
    if not data:
        _ex_factor_cache.update((ob, None) for ob in missing)
        return

    remain = set(r['order_book_id'] for r in data)
    # for each order_book_id, we add an init ex_factor value
    data.extend({'order_book_id': o, 'ex_date': MIN_DATE, 'ex_cum_factor': 1.0} for o in remain)

    df = pd.DataFrame(data)
    df = df[df.ex_date <= pd.datetime.today()]
    df.set_index('ex_date', inplace=True)
    df.sort_index(inplace=True)

    for order_book_id, s in df.groupby('order_book_id')['ex_cum_factor']:
        _ex_factor_cache[order_book_id] = s

    for order_book_id in missing:
        if order_book_id not in _ex_factor_cache:
            _ex_factor_cache[order_book_id] = None


def get_split_factor_for(order_book_ids, market):
    with _lock:
        _prepare_split_for(order_book_ids, market)
        return {ob: _split_factor_cache[ob] for ob in order_book_ids}


def _prepare_split_for(order_book_ids, market):
    global _ex_factor_cache, _expired, _split_factor_cache
    if _expired < time.time():
        _ex_factor_cache.clear()
        _split_factor_cache.clear()
        _expired = time.time() + 3600 * 8

    missing = [ob for ob in order_book_ids if ob not in _split_factor_cache]
    if not missing:
        return

    data = get_client().execute('_get_split', missing, market)
    if not data:
        _split_factor_cache.update((ob, None) for ob in missing)
        return

    remain = set(r['order_book_id'] for r in data)
    # for each order_book_id, we add an init split value
    data.extend(
        {
            'order_book_id': ob,
            'split_coefficient_to': 1.0,
            'split_coefficient_from': 1.0,
            'cum_factor': 1.0,
            'ex_dividend_date': MIN_DATE,
        } for ob in remain
    )

    df = pd.DataFrame(data)
    df = df[df.ex_dividend_date <= pd.datetime.today()]
    df.set_index('ex_dividend_date', inplace=True)
    df.sort_index(inplace=True)
    df["cum_factor"] = df["split_coefficient_to"] / df["split_coefficient_from"]
    df["cum_factor"] = df.groupby("order_book_id")["cum_factor"].cumprod()
    for order_book_id, s in df.groupby('order_book_id')['cum_factor']:
        s = s[~s.index.duplicated(keep='last')]
        # split index 有可能重复
        _split_factor_cache[order_book_id] = s.copy()

    for order_book_id in missing:
        if order_book_id not in _split_factor_cache:
            _split_factor_cache[order_book_id] = None


PRICE_FIELDS = {"low", "close", "limit_up", "limit_down", "unit_net_value", "open", "high"}

FIELDS_NEED_TO_ADJUST = PRICE_FIELDS | {'volume'}


def adjust_price_multi_df(df, order_book_ids, how, obid_slice_map, market):
    r_map_fields = {f: i for i, f in enumerate(df.columns) if f in FIELDS_NEED_TO_ADJUST}
    if not r_map_fields:
        # 没有需要复权的字段
        return
    # 是否前复权
    pre = how in ('pre', 'pre_volume')
    volume_adjust_by_ex_factor = how in ('pre_volume', 'post_volume')

    ex_factors = get_ex_factor_for(order_book_ids, market)
    volume_adjust_factors = {}

    if 'volume' in r_map_fields:
        if not volume_adjust_by_ex_factor:
            volume_adjust_factors = get_split_factor_for(order_book_ids, market)
        else:
            volume_adjust_factors = ex_factors

    data = df.values
    timestamps_level = df.index.get_level_values(1)
    for order_book_id, slice_ in obid_slice_map.items():
        if order_book_id not in order_book_ids:
            continue
        timestamps = timestamps_level[slice_]

        def calculate_factor(factors_map, order_book_id):
            factors = factors_map.get(order_book_id, None)
            if factors is not None:
                factor = np.take(factors.values, factors.index.searchsorted(timestamps, side='right') - 1)
                if pre:
                    factor /= factors.iloc[-1]
                return factor

        factor = calculate_factor(ex_factors, order_book_id)
        if factor is None:
            # 如果 ex_factor 为空，则不存在拆分及分红，无需后续处理
            continue

        if not volume_adjust_by_ex_factor:
            factor_volume = calculate_factor(volume_adjust_factors, order_book_id)
        else:
            factor_volume = factor

        for f, j in r_map_fields.items():
            if f in PRICE_FIELDS:
                data[slice_, j] *= factor
                np.round(data[slice_, j], 4, out=data[slice_, j])
            elif factor_volume is not None:
                data[slice_, j] *= 1 / factor_volume
                np.round(data[slice_, j], 0, out=data[slice_, j])



# coding: utf-8
from sqlalchemy import Numeric, Column
from .fundamental_base_sql import FundamentalBase


class AnaStkFinIdx(FundamentalBase):
    earnings_per_share = Column(Numeric(18, 4))
    fully_diluted_earnings_per_share = Column(Numeric(18, 4))
    diluted_earnings_per_share = Column(Numeric(18, 4))
    adjusted_earnings_per_share = Column(Numeric(18, 4))
    adjusted_fully_diluted_earnings_per_share = Column(Numeric(18, 4))
    adjusted_diluted_earnings_per_share = Column(Numeric(18, 4))
    book_value_per_share = Column(Numeric(18, 4))
    operating_cash_flow_per_share = Column(Numeric(18, 4))
    operating_total_revenue_per_share = Column(Numeric(18, 4))
    operating_revenue_per_share = Column(Numeric(18, 4))
    capital_reserve_per_share = Column(Numeric(18, 4))
    earned_reserve_per_share = Column(Numeric(18, 4))
    undistributed_profit_per_share = Column(Numeric(18, 4))
    retained_earnings_per_share = Column(Numeric(18, 4))
    cash_flow_from_operations_per_share = Column(Numeric(18, 4))
    ebit_per_share = Column(Numeric(18, 4))
    free_cash_flow_company_per_share = Column(Numeric(18, 4))
    free_cash_flow_equity_per_share = Column(Numeric(18, 4))
    dividend_per_share = Column(Numeric(18, 4))
    return_on_equity = Column(Numeric(18, 4))
    return_on_equity_weighted_average = Column(Numeric(18, 4))
    return_on_equity_diluted = Column(Numeric(18, 4))
    adjusted_return_on_equity_average = Column(Numeric(18, 4))
    adjusted_return_on_equity_weighted_average = Column(Numeric(18, 4))
    adjusted_return_on_equity_diluted = Column(Numeric(18, 4))
    return_on_asset = Column(Numeric(18, 4))
    return_on_asset_net_profit = Column(Numeric(18, 4))
    return_on_invested_capital = Column(Numeric(18, 4))
    annual_return_on_equity = Column(Numeric(18, 4))
    annual_return_on_asset = Column(Numeric(18, 4))
    annual_return_on_asset_net_profit = Column(Numeric(18, 4))
    net_profit_margin = Column(Numeric(18, 4))
    gross_profit_margin = Column(Numeric(18, 4))
    cost_to_sales = Column(Numeric(18, 4))
    net_profit_to_revenue = Column(Numeric(18, 4))
    profit_from_operation_to_revenue = Column(Numeric(18, 4))
    ebit_to_revenue = Column(Numeric(18, 4))
    expense_to_revenue = Column(Numeric(18, 4))
    operating_profit_to_profit_before_tax = Column(Numeric(18, 4))
    invesment_profit_to_profit_before_tax = Column(Numeric(18, 4))
    non_operating_profit_to_profit_before_tax = Column(Numeric(18, 4))
    income_tax_to_profit_before_tax = Column(Numeric(18, 4))
    adjusted_profit_to_total_profit = Column(Numeric(18, 4))
    debt_to_asset_ratio = Column(Numeric(18, 4))
    equity_multiplier = Column(Numeric(18, 4))
    current_asset_to_total_asset = Column(Numeric(18, 4))
    non_current_asset_to_total_asset = Column(Numeric(18, 4))
    interest_bearing_debt_to_capital = Column(Numeric(18, 4))
    current_debt_to_total_debt = Column(Numeric(18, 4))
    non_current_debt_to_total_debt = Column(Numeric(18, 4))
    current_ratio = Column(Numeric(18, 4))
    quick_ratio = Column(Numeric(18, 4))
    super_quick_ratio = Column(Numeric(18, 4))
    debt_to_equity_ratio = Column(Numeric(18, 4))
    equity_to_debt_ratio = Column(Numeric(18, 4))
    equity_to_interest_bearing_debt = Column(Numeric(18, 4))
    ebit_to_debt = Column(Numeric(18, 4))
    ocf_to_debt = Column(Numeric(18, 4))
    ocf_to_interest_bearing_debt = Column(Numeric(18, 4))
    ocf_to_current_ratio = Column(Numeric(18, 4))
    ocf_to_net_debt = Column(Numeric(18, 4))
    time_interest_earned_ratio = Column(Numeric(18, 4))
    long_term_debt_to_working_capital = Column(Numeric(18, 4))
    account_payable_turnover_rate = Column(Numeric(18, 4))
    account_payable_turnover_days = Column(Numeric(18, 4))
    account_receivable_turnover_days = Column(Numeric(18, 4))
    inventory_turnover = Column(Numeric(18, 4))
    account_receivable_turnover_rate = Column(Numeric(18, 4))
    current_asset_turnover = Column(Numeric(18, 4))
    fixed_asset_turnover = Column(Numeric(18, 4))
    total_asset_turnover = Column(Numeric(18, 4))
    inc_earnings_per_share = Column(Numeric(18, 4))
    inc_diluted_earnings_per_share = Column(Numeric(18, 4))
    inc_revenue = Column(Numeric(18, 4))
    inc_operating_revenue = Column(Numeric(18, 4))
    inc_gross_profit = Column(Numeric(18, 4))
    inc_profit_before_tax = Column(Numeric(18, 4))
    inc_net_profit = Column(Numeric(18, 4))
    inc_adjusted_net_profit = Column(Numeric(18, 4))
    inc_cash_from_operations = Column(Numeric(18, 4))
    inc_return_on_equity = Column(Numeric(18, 4))
    inc_book_per_share = Column(Numeric(18, 4))
    inc_total_asset = Column(Numeric(18, 4))
    du_return_on_equity = Column(Numeric(18, 4))
    du_equity_multiplier = Column(Numeric(18, 4))
    du_asset_turnover_ratio = Column(Numeric(18, 4))
    du_profit_margin = Column(Numeric(18, 4))
    du_return_on_sales = Column(Numeric(18, 4))
    non_recurring_profit_and_loss = Column(Numeric(18, 4))
    adjusted_net_profit = Column(Numeric(18, 4))
    ebit = Column(Numeric(18, 4))
    ebitda = Column(Numeric(18, 4))
    invested_capital = Column(Numeric(18, 4))
    working_capital = Column(Numeric(18, 4))
    net_working_capital = Column(Numeric(18, 4))
    retained_earnings = Column(Numeric(18, 4))
    interest_bearing_debt = Column(Numeric(18, 4))
    net_debt = Column(Numeric(18, 4))
    non_interest_bearing_current_debt = Column(Numeric(18, 4))
    non_interest_bearing_non_current_debt = Column(Numeric(18, 4))
    fcff = Column(Numeric(18, 4))
    fcfe = Column(Numeric(18, 4))
    depreciation_and_amortization = Column(Numeric(18, 4))
    ev = Column(Numeric(21, 4))
    ev_2 = Column(Numeric(21, 4))
    ev_to_ebit = Column(Numeric(18, 4))

    ev_to_ebitda = Column(Numeric(19, 4))
    tangible_assets = Column(Numeric(19, 4))
    tangible_asset_to_debt = Column(Numeric(19, 4))
    tangible_asset_to_interest_bearing_debt = Column(Numeric(19, 4))



# -*- coding: utf-8 -*-
import socket
import select

import warnings

from rqdatac.share.protocol import MSG_TYPE, SERIALIZATION_TYPE, COMPRESSION_METHOD, HEADER_LENGTH
from rqdatac.share.errors import get_error
from rqdatac.share.codec import snappy, brotli, msgpack, unpack_header, pack_one, unpack_one
from rqdatac.utils import connection_error

if brotli is not None:
    DEFAULT_COMPRESSION_METHOD = COMPRESSION_METHOD.BROTLI
elif snappy is not None:
    DEFAULT_COMPRESSION_METHOD = COMPRESSION_METHOD.SNAPPY
else:
    DEFAULT_COMPRESSION_METHOD = COMPRESSION_METHOD.ZLIB

if msgpack is None:
    DEFAULT_SERIALIZER_TYPE = SERIALIZATION_TYPE.JSON
else:
    DEFAULT_SERIALIZER_TYPE = SERIALIZATION_TYPE.MSGPACK


if hasattr(select, 'poll'):
    def _is_connection_normal(sock):
        poll = select.poll()
        poll.register(sock, select.POLLIN | select.POLLERR)
        return not poll.poll(0)
else:
    def _is_connection_normal(sock):
        readable, _, exceptional = select.select([sock], [], [sock], 0)
        return not readable and not exceptional


class ProtocolError(connection_error):
    pass


class Connection:
    sock_factory = staticmethod(socket.create_connection)

    @staticmethod
    def set_sock_factory(func):
        Connection.sock_factory = staticmethod(func)

    def __init__(self, sock, auth):
        self._socket = sock  # type: socket.socket
        self._sf = sock.makefile("rb", 128 * 1024)
        self._compression_method = DEFAULT_COMPRESSION_METHOD
        self._serializer_type = SERIALIZATION_TYPE.MSGPACK
        self._auth_info = auth.copy()
        self._auth_info["node"] = self._socket.getsockname()[0]
        self._do_auth(self._auth_info)

    def set_timeout(self, timeout):
        self._socket.settimeout(timeout)

    def is_normal(self):
        return _is_connection_normal(self._socket)

    def _read_one_packet(self):
        header = self._sf.read(HEADER_LENGTH)
        if len(header) != HEADER_LENGTH:
            if len(header) == 0:
                raise ProtocolError("Disconnected from the remote server")
            raise ProtocolError("incomplete header, %r" % len(header))

        mt, st, cm, length = unpack_header(header)
        if mt not in MSG_TYPE:
            raise ProtocolError("invalid message type: {}".format(mt))
        if not SERIALIZATION_TYPE.RAW <= st <= SERIALIZATION_TYPE.MSGPACK:
            raise ProtocolError("invalid serializer type: {}".format(st))
        if not COMPRESSION_METHOD.NONE <= cm <= COMPRESSION_METHOD.BROTLI:
            raise ProtocolError("invalid compression method: {}".format(cm))

        if length > 0:
            body = self._sf.read(length)
            if len(body) != length:
                raise ProtocolError("incomplete body")
            try:
                body = unpack_one(body, st, cm)
            except Exception as e:
                raise ProtocolError(repr(e))
            if mt == MSG_TYPE.ERROR:
                code, msg = body
                raise get_error(code)(msg)
            return mt, body

        return mt, None

    def _do_auth(self, auth):
        auth_body = pack_one(
            auth, MSG_TYPE.HANDSHAKE, DEFAULT_SERIALIZER_TYPE, DEFAULT_COMPRESSION_METHOD, force_compress=True
        )
        self._socket.sendall(auth_body)
        msg_type, body = self._read_one_packet()
        if msg_type != MSG_TYPE.HANDSHAKE:
            raise ProtocolError(
                "handshake message expected while {} received".format(MSG_TYPE[msg_type])
            )
        elif body is not None:
            warnings.warn(body)

    def close(self):
        self._sf.close()
        self._socket.close()

    def execute(self, method, *args, **kwargs):
        request_body = pack_one(
            (method, args, kwargs),
            MSG_TYPE.REQUEST,
            DEFAULT_SERIALIZER_TYPE,
            DEFAULT_COMPRESSION_METHOD,
        )
        self._socket.sendall(request_body)
        msg_type, body = self._read_one_packet()
        if msg_type == MSG_TYPE.RESPONSE:
            return body

        if msg_type != MSG_TYPE.STREAM_START:
            raise ProtocolError('got {} when STREAM_START expected'.format(MSG_TYPE[msg_type]))

        msg_type, body = self._read_one_packet()
        if msg_type == MSG_TYPE.STREAM_END:
            return []
        elif msg_type == MSG_TYPE.TABLE:
            return self.table_unpack(body)
        elif msg_type == MSG_TYPE.TABLE2:
            return self.table2_unpack(body)
        elif msg_type == MSG_TYPE.FEED:
            return self.feed_unpack(body)
        else:
            raise ProtocolError("unexpected message in stream: {}".format(MSG_TYPE[msg_type]))

    def feed_unpack(self, first):
        ret = list(first)
        while True:
            msg_type, body = self._read_one_packet()
            if msg_type == MSG_TYPE.STREAM_END:
                return ret
            elif msg_type == MSG_TYPE.FEED:
                ret.extend(body)
            else:
                raise ProtocolError(
                    "unexpected message in feed stream: {}".format(MSG_TYPE[msg_type])
                )

    def table_unpack(self, first):
        ret = []
        keys = first[0]
        ret.extend(dict(zip(keys, row)) for row in first[1:])
        while True:
            msg_type, body = self._read_one_packet()
            if msg_type == MSG_TYPE.STREAM_END:
                return ret
            elif msg_type == MSG_TYPE.TABLE:
                ret.extend(dict(zip(keys, row)) for row in body)
            else:
                raise ProtocolError(
                    "unexpected message in table stream: {}".format(MSG_TYPE[msg_type])
                )

    def table2_unpack(self, first):
        data = []
        head = first[0]
        data.extend(row for row in first[1:])
        while True:
            msg_type, body = self._read_one_packet()
            if msg_type == MSG_TYPE.STREAM_END:
                return head, data
            elif msg_type == MSG_TYPE.TABLE2:
                data.extend(row for row in body)
            else:
                raise ProtocolError(
                    "unexpected message in table2 stream: {}".format(MSG_TYPE[msg_type])
                )



# -*- coding: utf-8 -*-
import datetime
import socket

import pandas as pd
import numpy as np

from dateutil.parser import parse as parse_datetime
from dateutil.relativedelta import relativedelta

from six import string_types, integer_types, PY2, binary_type


def iterable(it):
    return hasattr(it, "__next__") or hasattr(it, "__iter__")


pd_version = pd.__version__
if pd_version >= "0.25":
    is_panel_removed = True
else:
    is_panel_removed = False

_str_like = string_types + (bytes,)


if PY2:
    connection_error = socket.error
    timeout_error = socket.timeout
else:
    timeout_error = TimeoutError
    connection_error = ConnectionError


def listify(it):
    if isinstance(it, _str_like):
        return [it]
    elif iterable(it):
        return list(it)
    else:
        return [it]


def to_datetime(dt):
    if isinstance(dt, datetime.datetime):
        return dt
    elif isinstance(dt, datetime.date):
        return datetime.datetime(dt.year, dt.month, dt.day)
    elif isinstance(dt, string_types):
        return parse_datetime(dt, ignoretz=True)
    elif isinstance(dt, integer_types):
        return int_to_datetime(dt)
    elif hasattr(dt, "to_pydatetime"):
        return dt.to_pydatetime()
    elif hasattr(dt, "dtype") and dt.dtype.char == "M":
        return parse_datetime(str(dt))
    raise ValueError("expect a datetime like object, got %r(%r)" % (type(dt), dt))


def to_date(dt):
    if isinstance(dt, datetime.datetime):
        return dt.date()
    elif isinstance(dt, datetime.date):
        return dt
    elif isinstance(dt, string_types):
        return parse_datetime(dt).date()
    elif isinstance(dt, integer_types):
        return int8_to_date(dt)
    elif hasattr(dt, "to_pydatetime"):
        return dt.to_pydatetime()
    elif hasattr(dt, "dtype") and dt.dtype.char == "M":
        return parse_datetime(str(dt)).date()
    raise ValueError("expect a datetime like object, got %r(%r)" % (type(dt), dt))


def to_datetime_str(dt):
    if not isinstance(dt, datetime.datetime):
        dt = to_datetime(dt)
    return "%04d-%02d-%02d %02d:%02d:%02d" % (
        dt.year,
        dt.month,
        dt.day,
        dt.hour,
        dt.minute,
        dt.second,
    )


def to_date_str(dt):
    if not isinstance(dt, datetime.date):
        dt = to_date(dt)
    return "%04d-%02d-%02d" % (dt.year, dt.month, dt.day)


def delay_today(years=0, months=0, days=0, leapdays=0, weeks=0):
    # type: (int, int, int, int, int) -> datetime.datetime
    return to_datetime(
        datetime.date.today()
        + relativedelta(years=years, months=months, days=days, leapdays=leapdays, weeks=weeks)
    )


def int_to_datetime(dt):
    # type: (int) -> datetime.datetime
    if 9999999 < dt < 99999999:  # 8位日期
        return int8_to_datetime(dt)
    if 9999999999999 < dt < 99999999999999:  # 14位日期时间
        return int14_to_datetime(dt)
    if 9999999999999999 < dt < 99999999999999999:  # 17位日期时间
        return int17_to_datetime(dt)
    raise ValueError("a datetime int should be 8, 14 or 17 length int, now is {}".format(dt))


def int8_to_datetime(dt):
    # type: (int) -> datetime.datetime
    year, dt = dt // 10000, dt % 10000
    month, day = dt // 100, dt % 100
    return datetime.datetime(year, month, day)


_int8_vectorize = np.vectorize(lambda y, m, d: datetime.datetime(y, m, d))


def int8_to_datetime_v(dtarr):
    if not isinstance(dtarr, np.ndarray):
        dtarr = np.array(dtarr)
    years, dt = dtarr // 10000, dtarr % 10000
    months, days = dt // 100, dt % 100
    return _int8_vectorize(years, months, days)


def int9_to_time(tm):
    hour, tm = tm // 10000000, tm % 10000000
    minute, tm = tm // 100000, tm % 100000
    second, ms = tm // 1000, tm % 1000
    return datetime.time(hour, minute, second, ms * 1000)


def int14_to_datetime(dt):
    # type: (int) -> datetime.datetime
    year, dt = dt // 10000000000, dt % 10000000000
    month, dt = dt // 100000000, dt % 100000000
    day, dt = dt // 1000000, dt % 1000000
    hour, dt = dt // 10000, dt % 10000
    minute, second = dt // 100, dt % 100
    return datetime.datetime(year, month, day, hour, minute, second)


_int14_vectorize = np.vectorize(lambda y, m, d, h, mm, s: datetime.datetime(y, m, d, h, mm, s))


def int14_to_datetime_v(dtarr):
    if not isinstance(dtarr, np.ndarray):
        dtarr = np.array(dtarr)
    years, dt = dtarr // 10000000000, dtarr % 10000000000
    months = dt // 100000000
    dt %= 100000000
    days = dt // 1000000
    dt %= 1000000
    hours = dt // 10000
    dt %= 10000
    minutes, seconds = dt // 100, dt % 100
    return _int14_vectorize(years, months, days, hours, minutes, seconds)


def int17_to_datetime(dt):
    # type: (int) -> datetime.datetime
    year, dt = dt // 10000000000000, dt % 10000000000000
    month, dt = dt // 100000000000, dt % 100000000000
    day, dt = dt // 1000000000, dt % 1000000000
    hour, dt = dt // 10000000, dt % 10000000
    minute, dt = dt // 100000, dt % 100000
    second, ms = dt // 1000, dt % 1000
    return datetime.datetime(year, month, day, hour, minute, second, ms * 1000)


_int17_vectorize = np.vectorize(lambda y, m, d, h, mm, s, ms: datetime.datetime(y, m, d, h, mm, s, ms))


def int17_to_datetime_v(dtarr):
    if not isinstance(dtarr, np.ndarray):
        dtarr = np.array(dtarr)
    years, dt = dtarr // 10000000000000, dtarr % 10000000000000
    months = dt // 100000000000
    dt %= 100000000000
    days = dt // 1000000000
    dt %= 1000000000
    hours = dt // 10000000
    dt %= 10000000
    minutes = dt // 100000
    dt %= 100000
    seconds, ms = dt // 1000, dt % 1000
    return _int17_vectorize(years, months, days, hours, minutes, seconds, ms*1000)


def int8_to_date(dt):
    # type: (int) -> datetime.date
    year, dt = dt // 10000, dt % 10000
    month, day = dt // 100, dt % 100
    return datetime.date(year, month, day)


def date_to_int8(dt):
    return dt.year * 10000 + dt.month * 100 + dt.day


def datetime_to_int14(dt):
    return (
        dt.year * 10000000000
        + dt.month * 100000000
        + dt.day * 1000000
        + dt.hour * 10000
        + dt.minute * 100
        + dt.second
    )


def datetime_to_int17(dt):
    return (
        dt.year * 10000000000000
        + dt.month * 100000000000
        + dt.day * 1000000000
        + dt.hour * 10000000
        + dt.minute * 100000
        + dt.second * 1000
        + int(round(dt.microsecond / 1000))  # ms have six digits
    )


def to_date_int(ds):
    # type: (...) -> int
    if isinstance(ds, int):
        return ds
    elif not isinstance(ds, (datetime.date, datetime.datetime)):
        ds = to_date(ds)
    year, month, day = ds.year, ds.month, ds.day
    return year * 10000 + month * 100 + day


def today_int():
    today = datetime.date.today()
    return today.year * 10000 + today.month * 100 + today.day


def str_to_dt_time(s):
    """ '21:31' -->  datetime.time(21, 31) """
    return datetime.time(int(s[0:2]), int(s[3:5]))


def safe_string_equal(s1, s2):
    if PY2:
        if isinstance(s1, binary_type):
            s1 = s1.decode("utf8")
        if isinstance(s2, binary_type):
            s2 = s2.decode("utf8")
    return s1 == s2


def pf_fill_nan(pf, order_book_ids):
    pf = pf.transpose(2, 0, 1)
    for order_book_id in order_book_ids:
        if order_book_id not in pf:
            pf[order_book_id] = np.NAN
    return pf.transpose(1, 2, 0)


def get_tick_value(tick, field, default=0):
    key_map = {"a": "ask", "b": "bid"}
    if field.startswith("a") or field.startswith("b"):
        key = key_map[field[0]]
        if field.endswith("v"):
            key += "_vol"
        t = tick.get(key)
        if t:
            return t[int(field[1]) - 1]
        else:
            return default
    else:
        return tick.get(field, default)


# -*- coding: utf-8 -*-

import bisect
import pandas as pd

from rqdatac.client import get_client
from rqdatac.decorators import export_as_api, compatible_with_parm
from rqdatac.validators import (
    ensure_date_int,
    ensure_date_range,
    ensure_list_of_string,
    ensure_order_book_ids,
    ensure_order_book_id
)
from rqdatac.utils import int8_to_datetime
from rqdatac.services.calendar import get_trading_dates_in_type


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def index_components(order_book_id, date=None, start_date=None, end_date=None, market="cn"):
    """获取指数成分
    :param order_book_id: 指数 id
    :param date: 指定日期；如不指定，返回最近一个交易日的数据
    :param start_date: 指定开始日期，不能和date同时指定
    :param end_date: 指定结束日期, 需和start_date同时指定并且应当不小于开始日期
    :param market:  (Default value = "cn")
    :returns list or dict
    """
    order_book_id = ensure_order_book_id(order_book_id)

    if date and (start_date or end_date):
        raise ValueError("date cannot be input together with start_date or end_date")
    elif (start_date and not end_date) or (end_date and not start_date):
        raise ValueError("start_date and end_date need to be applied together")

    if start_date:
        start_date, end_date = ensure_date_range(start_date, end_date)
        trading_dates = get_trading_dates_in_type(start_date, end_date, expect_type="int")
        if not trading_dates:
            return
        data = get_client().execute(
            "index_components_v2", order_book_id, trading_dates[0], trading_dates[-1], market=market)
        if not data:
            return
        data = {d["trade_date"]: d["component_ids"] for d in data}
        dates = sorted(data.keys())
        date0 = dates[0]
        res = {int8_to_datetime(dt): component_ids for dt, component_ids in data.items()}
        for trading_date in trading_dates:
            if trading_date < date0:
                continue
            if trading_date not in data:
                position = bisect.bisect_left(dates, trading_date) - 1
                res[int8_to_datetime(trading_date)] = data[dates[position]]
        return res

    if date:
        date = ensure_date_int(date)
    return get_client().execute("index_components", order_book_id, date, market=market)


@export_as_api
def index_weights(order_book_id, date=None, start_date=None, end_date=None, market="cn"):
    """获取指数的权重

    :param order_book_id: 如'000300.XSHG'和'000905.XSHG'
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :param date: 指定日期；如不指定，返回最近一个交易日的
    :param start_date: 指定开始日期，不能和date同时指定
    :param end_date: 指定结束日期, 需和start_date同时指定并且应当不小于开始日期
    :returns: 返回输入日期最近交易日该指数的权重

    """
    index_name = ensure_order_book_id(order_book_id)
    if date and (start_date or end_date):
        raise ValueError("date cannot be input together with start_date or end_date")
    elif (start_date and not end_date) or (end_date and not start_date):
        raise ValueError("start_date and end_date need to be applied together")

    if start_date:
        start_date, end_date = ensure_date_range(start_date, end_date)
        trading_dates = get_trading_dates_in_type(start_date, end_date, expect_type="int")
        if not trading_dates:
            return
        data = get_client().execute("index_weights_v2", index_name, trading_dates[0], trading_dates[-1], market=market)
        if not data:
            return

        data = {ensure_date_int(d["date"]): d["data"] for d in data}
        dates = sorted(data.keys())
        for trading_date in trading_dates:
            if trading_date not in data:
                position = bisect.bisect_left(dates, trading_date) - 1
                data[trading_date] = data[dates[position]]

        data = [
            {"date": int8_to_datetime(date), "order_book_id": c["order_book_id"], "weight": c["weight"]}
            for date, component_ids in data.items() if date in trading_dates for c in component_ids
        ]
        return pd.DataFrame(data).set_index(["date", "order_book_id"]).sort_index()

    if date:
        date = ensure_date_int(date)
    data = get_client().execute("index_weights", index_name, date, market=market)
    if not data:
        return
    s = pd.Series({d["order_book_id"]: d["weight"] for d in data})
    s.index.name = "order_book_id"
    return s


@export_as_api
def index_indicator(order_book_ids, start_date=None, end_date=None, fields=None, market="cn"):
    """获取指数指标

    :param order_book_ids: 如'000016.XSHG'
    :param start_date: 如'2016-01-01' (Default value = None)
    :param end_date: 如'2017-01-01' (Default value = None)
    :param fields: 如'pb', 默认返回全部 fields (Default value = None)
    :param market:  (Default value = "cn")
    :returns: pd.DataFrame 或 None

    """
    all_fields = ("pe_ttm", "pe_lyr", "pb_ttm", "pb_lyr", "pb_lf")
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if fields is not None:
        fields = ensure_list_of_string(fields)
        for f in fields:
            if f not in all_fields:
                raise ValueError("invalid field: {}".format(f))
    else:
        fields = all_fields

    df = get_client().execute(
        "index_indicator", order_book_ids, start_date, end_date, fields, market=market
    )
    if not df:
        return
    df = pd.DataFrame(df)
    df.set_index(["order_book_id", "trade_date"], inplace=True)
    return df



BALANCE_SHEET_FIELDS = [
    "deferred_expense",
    "dividend_payable",
    "short_term_debt",
    "financial_liabilities",
    "equity_parent_company",
    "tax_payable",
    "payroll_payable",
    "intangible_assets",
    "deferred_revenue",
    "total_assets",
    "depreciation_reserve",
    "accts_payable",
    "deferred_income_tax_liabilities",
    "other_fees_payable",
    "financial_asset_hold_to_maturity",
    "construction_in_progress",
    "goodwill",
    "other_payable",
    "other_payable_interest_dividend",
    "long_term_receivables",
    "accrued_expense",
    "capital_reserve",
    "accumulated_depreciation",
    "surplus_reserve",
    "real_estate_investment",
    "engineer_material",
    "minority_interest",
    "provision",
    "oil_and_gas_assets",
    "long_term_liabilities_due_one_year",
    "bond_payable",
    "net_fixed_assets",
    "cash_equivalent",
    "financial_asset_available_for_sale",
    "estimated_liabilities",
    "bad_debt_reserve",
    "other_non_current_liabilities",
    "non_current_liability_due_one_year",
    "non_current_assets",
    "long_term_deferred_expenses",
    "net_long_term_equity_investment",
    "bill_receivable",
    "non_current_asset_due_one_year",
    "total_equity_and_liabilities",
    "long_term_equity_investment",
    "total_liabilities",
    "prepaid_tax",
    "grants_received",
    "paid_in_capital",
    "housing_revolving_funds",
    "notes_payable",
    "long_term_payable",
    "other_non_current_assets",
    "interest_receivable",
    "impairment_intangible_assets",
    "advance_from_customers",
    "deferred_income_tax_assets",
    "other_current_assets",
    "fixed_asset_to_be_disposed",
    "financial_asset_held_for_trading",
    "short_term_loans",
    "inventory",
    "deferred_income",
    "interest_payable",
    "non_current_liabilities",
    "dividend_receivable",
    "other_current_liabilities",
    "current_assets",
    "long_term_loans",
    "capitalized_biological_assets",
    "current_liabilities",
    "contract_work",
    "total_equity",
    "prepayment",
    "net_accts_receivable",
    "undistributed_profit",
    "equity_preferred_stock",
    "other_accts_receivable",
    "other_equity_investment",
    "other_illiquidy_financial_assets",
    "life_reserve_receivable",
    "reinsurance_payable",
    "sub_issue_security_proceeds",
    "fixed_deposits",
    "insurance_receivable",
    "policy_dividend_payable",
    "mortgaged_loan",
    "comission_payable",
    "assets_hold_for_sale",
    "foreign_currency_converted_difference",
    "refundable_capital_deposits",
    "other_reserves",
    "loans_advances_to_customers",
    "perpetual_equity_debt",
    "debt_investment",
    "financial_lease_receivable",
    "uncertained_impairment_losses",
    "fund_providing",
    "unclaimed_reserve_receivable",
    "unearned_reserve_receivable",
    "preference_shares",
    "insurer_deposit_investment",
    "resale_financial_assets",
    "borrowings_capital",
    "other_composite_income",
    "deposits_from_interbank",
    "financial_lease_payable",
    "independent_account_assets",
    "bill_accts_receivable",
    "derivative_financial_liabilities",
    "derivative_financial_assets",
    "borrowings_from_central_banks",
    "buy_back_security_proceeds",
    "other_receivables_interest_dividend",
    "contract_assets",
    "reinsurance_receivable",
    "compensation_payable",
    "use_right_assets",
    "perpetual_bond",
    "loan_account_receivables",
    "health_reserve_receivable",
    "insurance_contract_reserve",
    "settlement_provision",
    "financial_receivable",
    "reinsurance_reserve_receivable",
    "other_liabilities",
    "unclaimed_indemnity_reserve",
    "treasury_stock",
    "subrogation_fee_receivable",
    "uncertained_premium_reserve",
    "client_provision",
    "refundable_deposits",
    "independent_account_liabilities",
    "liabilities_hold_for_sale",
    "life_insurance_reserve",
    "security_deposits_received",
    "trade_risk_allowances",
    "other_assets",
    "general_reserve",
    "other_equity_instruments",
    "deposits_of_interbank",
    "interbank_deposits",
    "deposits",
    "bill_accts_payable",
    "insurer_mortgage_loan",
    "specific_reserve",
    "lend_capital",
    "seat_costs",
    "client_deposits",
    "accrued_staff_costs",
    "health_insurance_reserve",
    "other_debt_investment",
    "precious_metals",
    "advance_insurance",
    "contract_liabilities",
    "lease_liabilities",
    "proxy_security_proceeds",
    "consumable_biological_assets",
    'raw_materials',
    'materials_in_transit',
    'development_product',
    'construction_product',
    'rental_development_product',
    'development_cost',
    'undeveloped_land',
    'commodity_stocks',
    'semi_finished_goods',
    'inprocess_goods',
    'finished_goods',
    'low_value_consumption_goods',
    'installment_sales',
    'goods_in_consignors',
    'goods_purchased',
    'fuel',
    'wrappage',
    'materials_in_consignors',
    'materials_purchased',
    'materials_cost_variance',
    'spare_parts',
    'flight_spare_parts',
    'expensive_turnover_parts',
    'common_equipment',
    'other_inventory',
    'total_fixed_assets',
    'total_construction_in_progress',
    'deposit_in_associate',
    'debt_assets',
    'debit_of_associate',
    'total_shares',
    'total_a_shares',
    'circulation_a_shares',
    'non_circulation_a_shares'
]

CASHFLOW_STATEMENT_FIELDS = [
    "exchange_rate_change_effect",
    "cash_paid_for_other_financing_activities",
    "cash_received_from_investors",
    "cash_from_operating_activities",
    "cash_paid_to_financing_activities",
    "fcff",
    "cash_paid_for_employee",
    "cash_paid_for_goods_and_services",
    "cash_from_other_operating_activities",
    "cash_received_from_disposal_of_asset",
    "cash_received_from_sales_of_goods",
    "deferred_expense_amortization",
    "cash_paid_for_other_investment_activities",
    "fcfe",
    "cash_flow_from_financing_activities",
    "cash_received_from_other_financing_activities",
    "cash_paid_for_dividend_and_interest",
    "cash_paid_for_debt",
    "fixed_asset_depreciation",
    "cash_paid_for_investment_activities",
    "cash_paid_for_asset",
    "cash_received_from_financing_activities",
    "cash_flow_from_investing_activities",
    "cash_received_from_investment_activities",
    "net_inc_cash_and_equivalents",
    "intangible_asset_amortization",
    "depreciation_and_amortization",
    "cash_paid_for_other_operation_activities",
    "cash_received_from_financial_institution_borrows",
    "cash_received_from_other_investment_activities",
    "cash_paid_to_acquire_investment",
    "cash_flow_from_operating_activities",
    "cash_equivalent_increase",
    "cash_paid_for_operation_activities",
    "cash_paid_for_taxes",
    "refunds_of_taxes",
    "assets_depreciation_reserves",
    "net_cash_deal_from_sub",
    "net_cash_payment_from_sub",
    "cash_received_from_interests_and_commissions",
    "cash_paid_for_orignal_insurance",
    "cash_paid_for_reinsurance",
    "net_increase_from_repurchasing_business",
    "net_increase_in_pledge_loans",
    "cash_received_from_disposal_of_investment",
    "cash_received_from_issuing_security",
    "cash_received_from_reinsurance",
    "cash_received_from_sub_issue_security",
    "net_increase_from_central_bank",
    "cash_received_from_minority_invest_subsidiaries",
    "net_increase_from_trading_financial_assets",
    "cash_received_from_investment",
    "net_increase_from_loans_and_advances",
    "cash_received_from_proxy_security",
    "net_increase_from_disposing_financial_assets",
    "end_period_cash_equivalent",
    "net_increase_from__financing_buy_back",
    "net_increase_from_insurer_deposit_investment",
    "cash_paid_for_comissions",
    "cash_received_from_original_insurance",
    "cash_received_from_issuing_equity_instruments",
    "dividends_paid_to_minority_by_subsidiaries",
    "net_increase_from_lending_capital",
    "cash_paid_for_policy_dividends",
    "net_increase_from_financial_institutions",
    "draw_back_canceled_loans",
    "other_effecting_cash_equivalent_items",
    "begin_period_cash_equivalent",
    "net_increase_from_central_bank_and_banks",
    "net_increase_from_operating_buy_back",
    "net_increase_from_investing_buy_back",
    "net_deposit_increase",
    "net_increase_from_other_financial_institutions",
]

INCOME_STATEMENT_FIELDS = [
    "total_income_minority",
    "ga_expense",
    "invest_income_associates",
    "profit_from_operation",
    "r_n_d",
    "total_income",
    "other_operating_revenue",
    "total_income_parent_company",
    "other_operating_cost",
    "revenue",
    "net_profit_parent_company",
    "non_operating_revenue",
    "selling_expense",
    "ebitda",
    "minority_profit",
    "disposal_loss_on_asset",
    "non_recurring_pnl",
    "exchange_gains_or_losses",
    "financing_expense",
    "interest_expense",
    "operating_revenue",
    "net_profit",
    "total_expense",
    "fair_value_change_income",
    "other_income",
    "asset_impairment",
    "non_operating_expense",
    "net_profit_deduct_non_recurring_pnl",
    "profit_before_tax",
    "sales_tax",
    "basic_earnings_per_share",
    "ebit",
    "investment_income",
    "income_tax",
    "gross_profit",
    "cost_of_goods_sold",
    "interest_income",
    "unrealised_investment_loss",
    "fully_diluted_earnings_per_share",
    "operating_expense",
    "o_n_a_expense",
    "credit_asset_impairment",
    "disposal_income_on_asset",
    "financing_interest_income",
    "other_debt_investment_change",
    "compensation_expense",
    "other_net_income",
    "reinsurance_income",
    "insurance_commission_expense",
    "corporate_credit_risk_change",
    "net_open_hedge_income",
    "other_effecting_net_profits_items",
    "remearsured_other_income",
    "other_revenue",
    "net_commission_income",
    "earned_premiums",
    "foreign_currency_statement_converted_difference",
    "classified_by_ownership",
    "other_income_classified_income_statement",
    "other_debt_investment_reserve",
    "discontinued_operation_net_profit",
    "premiums_income",
    "policy_dividend_payout",
    "other_income_equity_unclassified_income_statement",
    "sub_issue_security_income",
    "amortization_reinsurance_cost",
    "assets_reclassified_other_income",
    "commission_income",
    "continuous_operation_net_profit",
    "unearned_premium_reserve",
    "classified_by_continuity_operation",
    "other_effecting_total_profits_items",
    "net_interest_income",
    "cash_flow_hedging_effective_portion",
    "financial_asset_available_for_sale_change",
    "net_trust_income",
    "other_income_unclassified_income_statement",
    "other_income_parent_company",
    "other_income_equity_classified_income_statement",
    "net_proxy_security_income",
    "premium_reserve",
    "other_equity_instruments_change",
    "commission_expense",
    "refunded_premiums",
    "other_income_minority",
    "amortization_expense",
    "amortization_premium_reserve",
    "reinsurance_cost",
    "financial_asset_hold_to_maturity_change",
    "reinsurance",
    "others",
    "financing_interest_expense"
]

TTM_FIELDS = [
    "operating_expenseTTM",
    "operating_costTTM",
    "net_invest_cashflowTTM",
    "sale_service_render_cashTTM",
    "ebitTTM",
    "operating_revenueTTM",
    "operating_profitTTM",
    "ni_from_operatingTTM",
    "net_cashflowTTM",
    "net_finance_cashflowTTM",
    "net_operate_cashflowTTM",
    "operating_payoutTTM",
    "asset_impairment_lossTTM",
    "non_operating_net_incomeTTM",
    "administration_expenseTTM",
    "gross_profitTTM",
    "total_operating_costTTM",
    "financial_expenseTTM",
    "ni_from_value_changeTTM",
    "np_parent_company_ownersTTM",
    "net_profitTTM",
    "total_profitTTM",
    "total_operating_revenueTTM"]

MAINDATA_FIELDS = ["adjusted_net_profit"]

RND_EXPENDITURE_SHEET_FIELDS = [
    "expensed_rnd",
    "capitalized_rnd",
    "total_rnd",
    "rnd_to_revenue",
    "capitalized_rnd_ratio",
    "rnd_employees",
    "rnd_employees_ratio"]

FIELDS_LIST_EX = RND_EXPENDITURE_SHEET_FIELDS + MAINDATA_FIELDS + TTM_FIELDS + INCOME_STATEMENT_FIELDS + CASHFLOW_STATEMENT_FIELDS + BALANCE_SHEET_FIELDS
